from brdr.aligner import Aligner
from brdr.enums import AlignerInputType
from brdr.grb import update_to_actual_grb
from brdr.loader import GeoJsonFileLoader

aligner = Aligner()
# Load thematic data
loader = GeoJsonFileLoader("research.geojson", "OBJECTID")
aligner.load_thematic_data(loader)

fc = aligner.get_input_as_geojson(inputtype=AlignerInputType.THEMATIC)

fcs_actualisation = update_to_actual_grb(
    fc,
    id_theme_fieldname="OBJECTID",
    base_formula_field="brdr_formula",
    max_distance_for_actualisation=3,
    feedback=None,
)

print(fcs_actualisation)



import logging
from datetime import datetime
from enum import Enum

import requests
from shapely import box
from shapely.geometry import shape

from brdr.constants import DOWNLOAD_LIMIT, DEFAULT_CRS, DATE_FORMAT, VERSION_DATE
from brdr.loader import GeoJsonLoader
from brdr.logger import LOGGER
from brdr.utils import get_collection_by_partition

log = logging.getLogger(__name__)


class OEType(str, Enum):
    """
    Different types of Onroerend Eefgoed-objects are available:

    * AO: aanduidingsobjecten
    * EO: erfgoedobjecten
    """

    AO = "aanduidingsobjecten"
    EO = "erfgoedobjecten"


def get_oe_dict_by_ids(objectids, oetype=OEType.AO):
    """
    Fetches thematic data for a list of objectIDs from the Inventaris Onroerend Erfgoed
    API.

    This function retrieves information about designated heritage objects
    (erfgoedobjecten or aanduidingsobjecten) from the Flemish Agency for Heritage (
    Inventaris Onroerend Erfgoed) based on a list of their IDs.

    Args:
        objectids (list): A list of objectIDs of 'erfgoedobjecten' or
            'aanduidingsobjecten'.
        oetype (string): A string: 'aanduidingsobjecten' (default) or 'erfgoedobjecten'

    Returns:
        dict: A dictionary where keys are objectIDs (as strings) and values are
              GeoJSON geometry objects. If an erfgoedobject/aanduidingsobject is not
              found, a corresponding warning message will be logged, but it won't be\
              included in the returned dictionary.

    Raises:
        requests.exceptions.RequestException: If there is an error fetching data from
            the API.
    """
    logging.warning("deprecated method, use OnroerendErfgoedLoader instead")
    # TODO remove function
    dict_thematic = {}
    if oetype == OEType.AO:
        typename = "aanduidingsobjecten"
        # id_property = "aanduid_id"
    elif oetype == OEType.EO:
        typename = "erfgoedobjecten"
        # id_property = "erfgoed_id"
    else:
        logging.warning(
            "Undefined OE-type: " + str(oetype) + ": Empty collection returned"
        )
        return {}, None

    base_url = "https://inventaris.onroerenderfgoed.be/" + typename + "/"
    headers = {"Accept": "application/json"}
    for a in objectids:
        url = base_url + str(a)
        response = requests.get(url, headers=headers).json()
        if "id" in response.keys():
            key = str(response["id"])
            geom = shape(response["locatie"]["contour"])
            dict_thematic[key] = geom
        else:
            logging.warning("object id " + str(a) + " not available in " + oetype)
    return dict_thematic


def get_collection_oe_objects(
    oetype=OEType.AO,
    objectids=None,
    bbox=None,
    limit=DOWNLOAD_LIMIT,
    partition=1000,
    crs=DEFAULT_CRS,
):
    """
    Fetches GeoJSON data for designated heritage objects (aanduidingsobjecten) within
    a bounding box.

    This function retrieves information about aanduidingsobjecten from the Flemish
    Mercator public WFS service using a bounding box (bbox) as a filter. The bbox should
    be provided in the format "xmin,ymin,xmax,ymax" (EPSG:31370 projection).

    Args:
        bbox (str): A comma-separated string representing the bounding box in EPSG:31370
                   projection (e.g., "100000,500000,200000,600000").
        limit (int, optional): The maximum number of features to retrieve per request.
                              Defaults to 1000.

    Returns:
        dict: A dictionary containing the retrieved GeoJSON feature collection. This
              collection might be truncated if the total number of features exceeds
              the specified limit.
    """
    if oetype == OEType.AO:
        typename = "ps:ps_aandobj"
        id_property = "aanduid_id"
    elif oetype == OEType.EO:
        typename = "lu:lu_wet_erfgobj_pub"
        id_property = "erfgoed_id"
    else:
        logging.warning(
            "Undefined OE-type: " + str(oetype) + ": Empty collection returned"
        )
        return {}, None

    theme_url = (
        "https://www.mercator.vlaanderen.be/raadpleegdienstenmercatorpubliek/wfs?"
        "SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&"
        f"TYPENAMES={typename}&"
        f"SRSNAME={crs}"
        "&outputFormat=application/json"
    )
    if objectids is not None:
        filter = (
            f"&CQL_FILTER={id_property} IN ("
            + ", ".join(str(o) for o in objectids)
            + ")"
        )
        theme_url = theme_url + filter
    bbox_polygon = None
    if bbox is not None:
        bbox_polygon = box(*tuple(o for o in bbox))

    return (
        get_collection_by_partition(
            theme_url, geometry=bbox_polygon, partition=partition, limit=limit, crs=crs
        ),
        id_property,
    )


class OnroerendErfgoedLoader(GeoJsonLoader):
    def __init__(
        self,
        objectids=None,
        oetype=OEType.AO,
        bbox=None,
        limit=DOWNLOAD_LIMIT,
        partition=1000,
        crs=DEFAULT_CRS,
    ):
        if (objectids is None and bbox is None) or (
            objectids is not None and bbox is not None
        ):
            raise ValueError("Please provide a ID-filter OR a BBOX-filter, not both")
        super().__init__()
        self.objectids = objectids
        self.oetype = oetype
        self.bbox = bbox
        self.limit = limit
        self.part = partition
        self.crs = crs
        self.data_dict_source["source"] = "Onroerend Erfgoed"

    def load_data(self):

        # geom_union = buffer_pos(self.aligner.get_thematic_union(), MAX_REFERENCE_BUFFER)
        collection, id_property = get_collection_oe_objects(
            oetype=self.oetype,
            objectids=self.objectids,
            bbox=self.bbox,
            partition=self.part,
            limit=self.limit,
            crs=self.crs,
        )
        self.id_property = id_property
        self.input = dict(collection)
        self.data_dict_source[VERSION_DATE] = datetime.now().strftime(DATE_FORMAT)
        LOGGER.debug(f"OnroerendErfgoed-objects downloaded")
        return super().load_data()



import numpy as np

from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import GRBActualLoader
from brdr.loader import GeoJsonFileLoader
from brdr.utils import diffs_from_dict_series
from examples import plot_series

if __name__ == "__main__":
    """
    # example to check if we can notice if it is better to align to a building instead of a
    # parcel
    """
    # Initiate brdr
    aligner_x = Aligner()
    # Load thematic data & reference data (parcels)
    aligner_x.load_thematic_data(
        GeoJsonFileLoader(
            "../tests/testdata/test_parcel_vs_building.geojson", "theme_id"
        )
    )
    aligner_x.load_reference_data(
        GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner_x)
    )  # gebruik de actuele adp-percelen adp= administratieve percelen

    aligner_y = Aligner()
    # Load thematic data & reference data (buildings)
    aligner_y.load_thematic_data(
        GeoJsonFileLoader(
            "../tests/testdata/test_parcel_vs_building.geojson", "theme_id"
        )
    )
    aligner_y.load_reference_data(
        GRBActualLoader(grb_type=GRBType.GBG, partition=1000, aligner=aligner_y)
    )  # gebruik de actuele adp-percelen adp= administratieve percelen

    # Example how to use a series (for histogram)
    series = np.arange(0, 310, 10, dtype=int) / 100
    x_dict_series = aligner_x.process(
        relevant_distances=series, od_strategy=4, threshold_overlap_percentage=50
    )
    x_resulting_areas = diffs_from_dict_series(x_dict_series, aligner_x.dict_thematic)
    y_dict_series = aligner_y.process(
        relevant_distances=series, od_strategy=4, threshold_overlap_percentage=50
    )
    y_resulting_areas = diffs_from_dict_series(y_dict_series, aligner_y.dict_thematic)
    # plot_diffs(series,x_resulting_areas)
    # plot_diffs(series,y_resulting_areas)

    # Make a 1-by-1 comparison for each thematic feature compared to the 2 references (
    # parcels and buildings)
    for key in x_resulting_areas:
        dict_diff = {
            "x" + str(key): x_resulting_areas[key],
            "y" + str(key): y_resulting_areas[key],
        }
        plot_series(series, dict_diff)



from brdr.aligner import Aligner
from brdr.enums import OpenbaarDomeinStrategy, GRBType
from brdr.grb import GRBActualLoader
from brdr.loader import GeoJsonLoader
from brdr.utils import diffs_from_dict_series
from examples import plot_series
from examples import show_map

if __name__ == "__main__":
    # EXAMPLE to process a series of relevant distances
    # Initiate brdr
    aligner = Aligner()
    # Load thematic data
    thematic_json = {
        "type": "FeatureCollection",
        "name": "test",
        "crs": {"type": "name", "properties": {"name": "urn:ogc:def:crs:EPSG::31370"}},
        "features": [
            {
                "type": "Feature",
                "properties": {"fid": 1100, "id": 1100, "theme_identifier": "1100"},
                "geometry": {
                    "type": "MultiPolygon",
                    "coordinates": [
                        [
                            [
                                [170020.885142877610633, 171986.324472956912359],
                                [170078.339491307124263, 172031.344329243671382],
                                [170049.567976467020344, 172070.009247593494365],
                                [170058.413533725659363, 172089.43287940043956],
                                [170071.570170061604585, 172102.403589786874363],
                                [170061.212614970601862, 172153.235667688539252],
                                [170008.670597244432429, 172137.344562214449979],
                                [169986.296310421457747, 172121.231194059771951],
                                [169979.658902874827618, 172095.676061166130239],
                                [169980.509304589155363, 172078.495172578957863],
                                [169985.424311356124235, 172065.162689057324314],
                                [169971.609697333740769, 172057.093288242496783],
                                [170020.885142877610633, 171986.324472956912359],
                            ]
                        ]
                    ],
                },
            }
        ],
    }

    loader = GeoJsonLoader(_input=thematic_json, id_property="theme_identifier")
    aligner.load_thematic_data(loader)
    # Load reference data: The actual GRB-parcels
    aligner.load_reference_data(
        GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner)
    )
    # PROCESS a series of relevant distances
    relevant_distances = [0.5, 1, 3, 6]
    dict_results = aligner.process(
        relevant_distances=relevant_distances,
        od_strategy=OpenbaarDomeinStrategy.SNAP_ALL_SIDE,
        threshold_overlap_percentage=50,
    )
    # SHOW results: map and plotted changes
    show_map(dict_results, aligner.dict_thematic, aligner.dict_reference)
    resulting_areas = diffs_from_dict_series(dict_results, aligner.dict_thematic)
    plot_series(relevant_distances, resulting_areas)



import json
import os

import pytest
import responses

here = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(autouse=True, scope="session")
def requests_mock():
    """Block all requests calls."""
    with responses.RequestsMock() as rsps:
        rsps.add_passthru("https://www.mercator.vlaanderen.be")
        rsps.add_passthru("https://geo.api.vlaanderen.be")
        rsps.add_passthru("https://inventaris.onroerenderfgoed.be")
        yield rsps


@pytest.fixture(scope="session")
def haspengouw_geojson():
    with open(f"{here}/testdata/haspengouw.geojson", "r") as f:
        return json.loads(f.read())



from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import get_collection_grb_actual, GRBActualLoader
from brdr.loader import GeoJsonFileLoader, DictLoader
from brdr.utils import polygonize_reference_data, geojson_to_dicts
from examples import show_map, print_brdr_formula


if __name__ == "__main__":
    """
    # example to test what happens if we combine borders
    # (so thematic data can use both polygons)

    # If we just combine sets of polygons (fe parcels and buildings) these polygons will
    # overlap, and gives some unwanted/unexpected results. The reference data can be
    # preprocessed with 'shapely-polygonize', to create a new set of non-overlapping
    # polygons. These non-overlapping polygons can be used as reference-data to align the
    # theme-data.
    """
    # Initiate brdr
    aligner = Aligner()

    # Load thematic data & reference data
    loader = GeoJsonFileLoader(
        path_to_file="../tests/testdata/test_parcel_vs_building.geojson",
        id_property="theme_id",
    )
    aligner.load_thematic_data(loader)

    adploader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner)
    gbgloader = GRBActualLoader(grb_type=GRBType.GBG, partition=1000, aligner=aligner)
    collection_adp, name_reference_id_adp = get_collection_grb_actual(
        aligner.get_thematic_union(),
        grb_type=GRBType.ADP,
        partition=1000,
        date_start=None,
        date_end=None,
    )
    dict_adp, dict_adp_properties = geojson_to_dicts(
        collection_adp, name_reference_id_adp
    )

    collection_gbg, name_reference_id_gbg = get_collection_grb_actual(
        aligner.get_thematic_union(),
        grb_type=GRBType.GBG,
        partition=1000,
        date_start=None,
        date_end=None,
    )
    dict_gbg, dict_gbg_properties = geojson_to_dicts(
        collection_gbg, name_reference_id_gbg
    )

    dict_adp_gbg = dict_adp
    dict_adp_gbg.update(dict_gbg)  # combine 2 dictionaries
    # make a polygonized version of the reference data with non-overlapping polygons
    dict_ref = polygonize_reference_data(dict_adp_gbg)
    aligner.load_reference_data(DictLoader(dict_ref))

    rel_dist = 2
    dict_results = aligner.process(relevant_distances=[rel_dist], od_strategy=4)
    aligner.save_results("output/")
    show_map(dict_results, aligner.dict_thematic, aligner.dict_reference)
    print_brdr_formula(dict_results, aligner)



import numpy as np

from brdr.aligner import Aligner
from brdr.enums import GRBType, AlignerResultType, OpenbaarDomeinStrategy
from brdr.grb import GRBActualLoader
from brdr.loader import GeoJsonFileLoader
from examples import show_map, plot_series

# Press the green button in the gutter to run the script.
if __name__ == "__main__":
    """
    EXAMPLE to use the predictor-function to automatically predict which resulting
    geometries are interesting to look at (based on detection of breakpoints and
    relevant distances of 'no-change')
    """
    # Initiate an Aligner
    aligner = Aligner()
    # Load thematic data & reference data
    loader = GeoJsonFileLoader(
        "../tests/testdata/test_wanted_changes.geojson", "theme_id"
    )
    aligner.load_thematic_data(loader)
    # Load reference data
    loader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner)
    aligner.load_reference_data(loader)

    # PREDICT the 'stable' relevant distances, for a series of relevant distances
    series = np.arange(0, 310, 10, dtype=int) / 100
    # predict which relevant distances are interesting to propose as resulting geometry
    dict_series, dict_predictions, diffs = aligner.predictor(
        relevant_distances=series,
        od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE,
        threshold_overlap_percentage=50,
    )

    # SHOW results of the predictions
    fcs = aligner.get_results_as_geojson(
        resulttype=AlignerResultType.PREDICTIONS, formula=False
    )
    print(fcs["result"])
    for key in dict_predictions:
        plot_series(series, {key: diffs[key]})
        show_map(
            {key: dict_predictions[key]},
            {key: aligner.dict_thematic[key]},
            aligner.dict_reference,
        )



# define a typeddict thematic_data with keys name: str and geom: geometry
from typing import Dict
from typing import List
from typing import TypedDict

from shapely.geometry.base import BaseGeometry


class GeoJSONGeometry(TypedDict):
    type: str
    coordinates: List


class Crs(TypedDict):
    type: str
    properties: dict


class Feature(TypedDict):
    type: str
    geometry: GeoJSONGeometry
    properties: dict


class FeatureCollection(TypedDict, total=False):
    type: str
    name: str
    crs: Crs
    features: List[Feature]
    __extra_items__: Dict[str, str]


class ProcessResult(TypedDict, total=False):
    result: BaseGeometry
    result_diff: BaseGeometry
    result_diff_plus: BaseGeometry
    result_diff_min: BaseGeometry
    result_relevant_intersection: BaseGeometry
    result_relevant_diff: BaseGeometry
    remark: str



# Download-settings: when extracting features by URL
# Limit used when extracting features by URL, using the feature API (f.e. from GRB)
DOWNLOAD_LIMIT = 10000
# default CRS:
DEFAULT_CRS = "EPSG:31370"  # BelgianLambert72

# MULTI_SINGLE_ID_SEPARATOR #separator to split multipolygon_ids to single polygons
MULTI_SINGLE_ID_SEPARATOR = "*$*"

PREFIX_FIELDNAME = "brdr_"
BASE_FORMULA_FIELD_NAME = (
    PREFIX_FIELDNAME + "base_formula"
)  # for use in grb_actualisation
FORMULA_FIELD_NAME = PREFIX_FIELDNAME + "formula"
EVALUATION_FIELD_NAME = PREFIX_FIELDNAME + "evaluation"
DIFF_PERCENTAGE_FIELD_NAME = PREFIX_FIELDNAME + "diff_percentage"
DIFF_AREA_FIELD_NAME = PREFIX_FIELDNAME + "diff_area"
FULL_BASE_FIELD_NAME = PREFIX_FIELDNAME + "full_base"
FULL_ACTUAL_FIELD_NAME = PREFIX_FIELDNAME + "full_actual"
EQUAL_REFERENCE_FEATURES_FIELD_NAME = PREFIX_FIELDNAME + "equal_reference_features"
OD_ALIKE_FIELD_NAME = PREFIX_FIELDNAME + "od_alike"

AREA_ATTRIBUTE = PREFIX_FIELDNAME + "area"
PERIMETER_ATTRIBUTE = PREFIX_FIELDNAME + "perimeter"
SHAPE_INDEX_ATTRIBUTE = PREFIX_FIELDNAME + "shape_index"

NR_CALCULATION_FIELD_NAME = PREFIX_FIELDNAME + "nr_calculations"
RELEVANT_DISTANCE_FIELD_NAME = PREFIX_FIELDNAME + "relevant_distance"
REMARK_FIELD_NAME = PREFIX_FIELDNAME + "remark"
LAST_VERSION_DATE = "last_version_date"
VERSION_DATE = "version_date"
DATE_FORMAT = "%Y-%m-%d"

# GRB_CONSTANTS
# max buffer (m) around thematic geometry to download reference parcels
GRB_MAX_REFERENCE_BUFFER = 10
# URL of the OGC feature API of actual GRB to extract collections
GRB_FEATURE_URL = "https://geo.api.vlaanderen.be/GRB/ogc/features/collections"
# URL of the OGC feature API of GRB fiscal parcels (situation of 1st of January) to
# extract collections
GRB_FISCAL_PARCELS_URL = "https://geo.api.vlaanderen.be/Adpf/ogc/features/collections"
# Property-name of version_date
GRB_VERSION_DATE = "VERSDATUM"
# Property-name of id of GRB-parcels
GRB_PARCEL_ID = "CAPAKEY"
# Property-name of id of GRB-parcels
GRB_BUILDING_ID = "OIDN"
# Property-name of id of GRB-parcels
GRB_KNW_ID = "OIDN"



import unittest

import shapely
from shapely import is_empty
from shapely.geometry import Polygon

from brdr.constants import MULTI_SINGLE_ID_SEPARATOR
from brdr.oe import get_oe_dict_by_ids
from brdr.typings import ProcessResult
from brdr.utils import diffs_from_dict_series

# from brdr.utils import filter_dict_by_key
from brdr.utils import get_breakpoints_zerostreak
from brdr.utils import get_collection
from brdr.utils import merge_process_results
from brdr.utils import multipolygons_to_singles
from brdr.utils import polygonize_reference_data


class TestUtils(unittest.TestCase):
    def setUp(self):
        # Create a sample geometry for testing
        self.sample_series = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    def tearDown(self):
        pass

    def test_get_breakpoints_zerostreak(self):
        sample_diff = [0, 5, 5, 3, 2, 2, 2, 6, 7, 7, 6]
        breakpoints, zerostreaks = get_breakpoints_zerostreak(
            self.sample_series, sample_diff
        )
        assert len(breakpoints) != 0
        assert len(zerostreaks) != 0

    def test_get_breakpoints_zerostreak_no_zerostreaks(self):
        breakpoints, zerostreaks = get_breakpoints_zerostreak(
            self.sample_series, self.sample_series
        )
        assert len(breakpoints) != 0
        assert len(zerostreaks) == 0

    def test_multipolygons_to_singles_empty_dict(self):
        data = {}
        result, dict_multi_as_single = multipolygons_to_singles(data)
        self.assertEqual(result, {})

    def test_multipolygons_to_singles_with_point(self):
        geometry1 = shapely.geometry.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        geometry2 = shapely.geometry.Point(0, 0)
        data = {"test_id1": geometry1, "test_id2": geometry2}
        result, dict_multi_as_single = multipolygons_to_singles(data)
        self.assertEqual(result, {"test_id1": geometry1})

    def test_multipolygons_to_singles_single_polygon(self):
        geometry = shapely.geometry.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        data = {"test_id": geometry}
        result, dict_multi_as_single = multipolygons_to_singles(data)
        self.assertEqual(result, data)

    def test_multipolygons_to_singles_multipolygon_single_poly(self):
        geometry = shapely.geometry.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        data = {"test_id": shapely.geometry.MultiPolygon([geometry])}
        result, dict_multi_as_single = multipolygons_to_singles(data)
        self.assertEqual(result, {"test_id": geometry})

    def test_polygonize_reference_data_no_overlap(self):
        """Tests polygonize_reference_data with non-overlapping polygons."""
        data = {
            "ref1": Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
            "ref2": Polygon([(2, 2), (3, 2), (3, 3), (2, 3)]),
        }
        result = polygonize_reference_data(data.copy())
        # Assert expected number of features and new keys
        self.assertEqual(len(result), 2)
        self.assertIn("1", result.keys())
        self.assertIn("2", result.keys())

    def test_polygonize_reference_data_simple_overlap(self):
        """Tests polygonize_reference_data with a simple overlapping case."""
        data = {
            "ref1": Polygon([(0, 0), (2, 0), (2, 1), (1, 1), (0, 0.5)]),
            "ref2": Polygon([(1, 0), (3, 0), (3, 2), (1, 2)]),
        }
        result = polygonize_reference_data(data.copy())
        # Assert expected number of features (might be more than original due to
        # splitting)
        self.assertGreaterEqual(len(result), 2)
        # Assert new keys are used
        for key in result.keys():
            self.assertTrue(key.isdigit())

    def test_get_oe_dict_by_ids(self):
        aanduid_id = 131635
        dict_thematic = get_oe_dict_by_ids([aanduid_id])
        self.assertFalse(is_empty(dict_thematic[str(aanduid_id)]))

    # def test_get_oe_dict_by_ids_erfgoedobject(self):
    #     eo_id = 206363
    #     dict_thematic = get_oe_dict_by_ids([eo_id], oetype=OEType.EO)
    #     self.assertFalse(is_empty(dict_thematic[str(eo_id)]))
    #
    # def test_get_oe_dict_by_ids_empty(self):
    #     dict_thematic = get_oe_dict_by_ids([])
    #     self.assertEqual(dict_thematic, {})
    #
    # def test_get_oe_dict_by_ids_not_existing(self):
    #     aanduid_id = -1
    #     dict_thematic = get_oe_dict_by_ids([aanduid_id])
    #     self.assertEqual(dict_thematic, {})

    # def test_filter_dict_by_key_empty_dict(self):
    #     data = {}
    #     result = filter_dict_by_key(data, "key")
    #     self.assertEqual(result, {})
    #
    # def test_filter_dict_by_key_single_match(self):
    #     data = {"key1": "value1", "key2": "value2"}
    #     result = filter_dict_by_key(data, "key1")
    #     self.assertEqual(result, {"key1": "value1"})
    #
    # def test_filter_dict_by_key_no_match(self):
    #     data = {"key1": "value1", "key2": "value2"}
    #     result = filter_dict_by_key(data, "key3")
    #     self.assertEqual(result, {})

    def test_diffs_from_dict_series_complete(self):
        """Tests diffs_from_dict_series with complete data."""
        # Mock data
        dict_thematic = {
            "theme_id1": Polygon([(0, 0), (10, 0), (10, 10), (0, 10)]),
            "theme_id2": Polygon([(5, 5), (15, 5), (15, 15), (5, 15)]),
        }
        dict_series = {
            "theme_id1": {
                10: {
                    "result": Polygon([(0, 0), (8, 0), (8, 8), (0, 8)]),
                    "result_diff": Polygon([(2, 2), (6, 2), (6, 6), (2, 6)]),
                }
            },
            "theme_id2": {
                10: {
                    "result": Polygon([(7, 7), (13, 7), (13, 13), (7, 13)]),
                    "result_diff": Polygon([(9, 9), (11, 9), (11, 11), (9, 11)]),
                }
            },
        }
        expected_diffs = {"theme_id1": {10: 16.0}, "theme_id2": {10: 4.0}}

        assert expected_diffs == diffs_from_dict_series(
            dict_series.copy(), dict_thematic.copy()
        )

    def test_get_collection(self):
        base_year = 2023
        limit = 100
        crs = "EPSG:31370"
        bbox = "173500,173500,174000,174000"
        ref_url = (
            "https://geo.api.vlaanderen.be/Adpf/ogc/features/collections/Adpf"
            + str(base_year)
            + "/items?"
            "limit=" + str(limit) + "&crs=" + crs + "&bbox-crs=EPSG:31370&bbox=" + bbox
        )
        collection = get_collection(ref_url, limit)
        self.assertTrue("features" in collection.keys())

    def test_merge_process_results(self):
        key_1 = "key" + MULTI_SINGLE_ID_SEPARATOR + "1"
        key_2 = "key" + MULTI_SINGLE_ID_SEPARATOR + "2"
        key_3 = "key_3"
        dict_multi_as_single = {key_1: "key", key_2: "key"}
        process_result_1 = ProcessResult()
        process_result_1["result"] = Polygon([(0, 0), (10, 0), (10, 10), (0, 10)])
        process_result_2 = ProcessResult()
        process_result_2["result"] = Polygon([(0, 0), (8, 0), (8, 8), (0, 8)])
        process_result_3 = ProcessResult()
        process_result_3["result"] = Polygon([(0, 0), (8, 0), (8, 8), (0, 8)])
        testdict = {
            key_1: {0: process_result_1},
            key_2: {0: process_result_2},
            key_3: {0: process_result_3},
        }
        merged_testdict = merge_process_results(
            result_dict=testdict, dict_multi_as_single=dict_multi_as_single
        )
        assert len(merged_testdict.keys()) == 2



import logging
import os.path

import numpy as np
import requests
from geojson import Feature, FeatureCollection, dump
from shapely import GeometryCollection, make_valid, node, polygonize, unary_union
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry

from brdr.constants import (
    MULTI_SINGLE_ID_SEPARATOR,
    DEFAULT_CRS,
    DOWNLOAD_LIMIT,
    RELEVANT_DISTANCE_FIELD_NAME,
    NR_CALCULATION_FIELD_NAME,
    REMARK_FIELD_NAME,
    PERIMETER_ATTRIBUTE,
    SHAPE_INDEX_ATTRIBUTE,
    AREA_ATTRIBUTE,
)
from brdr.enums import DiffMetric
from brdr.geometry_utils import get_partitions, get_bbox
from brdr.typings import ProcessResult


def get_series_geojson_dict(
    series_dict: dict[any, dict[float, ProcessResult]],
    crs: str,
    id_field: str,
    series_prop_dict: dict[any, dict[float, any]] = None,
    geom_attributes=True,
):
    """
    Convert a series of process results to a GeoJSON feature collection.

    Args:
        series_dict (dict): Dictionary containing process results.
        crs (str): Coordinate reference system.
        id_field (str): Field name for the ID.
        series_prop_dict (dict, optional): Dictionary containing series properties.
        geom_attributes (bool, optional): Whether to include geometry attributes.

    Returns:
        dict: Dictionary of GeoJSON feature collections.
    """
    features_list_dict = {}

    for theme_id, results_dict in series_dict.items():
        nr_calculations = len(results_dict)
        prop_dict = dict(series_prop_dict or {}).get(theme_id, {})
        for relative_distance, process_result in results_dict.items():
            properties = prop_dict.get(relative_distance, {})
            properties[id_field] = theme_id
            properties[NR_CALCULATION_FIELD_NAME] = nr_calculations
            properties[RELEVANT_DISTANCE_FIELD_NAME] = relative_distance
            if "remark" in process_result:
                properties[REMARK_FIELD_NAME] = process_result["remark"]

            for results_type, geom in process_result.items():
                if not isinstance(geom, BaseGeometry):
                    continue
                if results_type not in features_list_dict:
                    features_list_dict[results_type] = []

                feature = _feature_from_geom(geom, properties, geom_attributes)
                features_list_dict[results_type].append(feature)

    crs_geojson = {"type": "name", "properties": {"name": crs}}
    return {
        result_type: FeatureCollection(features, crs=crs_geojson)
        for result_type, features in features_list_dict.items()
    }


def _feature_from_geom(
    geom: BaseGeometry,
    properties: dict = None,
    geom_attributes=True,
) -> Feature:
    """
    Convert a geometry to a GeoJSON feature.

    Args:
        geom (BaseGeometry): The geometry to convert.
        properties (dict, optional): The properties to include in the feature.
        geom_attributes (bool, optional): Whether to include geometry attributes.

    Returns:
        Feature: The GeoJSON feature.
    """
    properties = dict(properties or {})
    if geom_attributes:
        area = geom.area
        perimeter = geom.length
        properties[AREA_ATTRIBUTE] = area
        properties[PERIMETER_ATTRIBUTE] = perimeter
        properties[SHAPE_INDEX_ATTRIBUTE] = perimeter / area if area != 0 else -1
    return Feature(geometry=geom, properties=properties)


def geojson_from_dict(dictionary, crs, id_field, prop_dict=None, geom_attributes=True):
    """
    Get a GeoJSON (FeatureCollection) from a dictionary of IDs (keys) and geometries (values).

    Args:
        dictionary (dict): Dictionary of geometries.
        crs (str): Coordinate reference system.
        id_field (str): Field name for the ID.
        prop_dict (dict, optional): Dictionary of properties.
        geom_attributes (bool, optional): Whether to include geometry attributes.

    Returns:
        FeatureCollection: The GeoJSON FeatureCollection.
    """
    features = []
    for key, geom in dictionary.items():
        properties = dict(prop_dict or {}).get(key, {})
        properties[id_field] = key
        features.append(_feature_from_geom(geom, properties, geom_attributes))
    crs_geojson = {"type": "name", "properties": {"name": crs}}
    geojson = FeatureCollection(features, crs=crs_geojson)
    return geojson


def write_geojson(path_to_file, geojson):
    """
    Write a GeoJSON object to a file.

    Args:
        path_to_file (str): Path to the output file.
        geojson (FeatureCollection): The GeoJSON object to write.
    """
    parent = os.path.dirname(path_to_file)
    os.makedirs(parent, exist_ok=True)
    with open(path_to_file, "w") as f:
        dump(geojson, f, default=str)


def multipolygons_to_singles(dict_geoms):
    """
    Convert a dictionary of Shapely geometries to a dictionary containing only single polygons.

    Args:
        dict_geoms (dict): Dictionary of geometries.

    Returns:
        tuple: A tuple containing:
            - dict: Dictionary of single polygons.
            - dict: Dictionary mapping new keys to original keys.
    """
    resulting_dict_geoms = {}
    dict_multi_as_single = {}
    for key, geom in dict_geoms.items():
        if str(geom.geom_type) == "Polygon":
            resulting_dict_geoms[key] = geom
        elif str(geom.geom_type) == "MultiPolygon":
            polygons = list(geom.geoms)
            if len(polygons) == 1:
                resulting_dict_geoms[key] = polygons[0]
                continue
            i = 0
            for p in polygons:
                new_key = str(key) + MULTI_SINGLE_ID_SEPARATOR + str(i)
                dict_multi_as_single[new_key] = key
                resulting_dict_geoms[new_key] = p
                i = i + 1
        else:
            logging.debug("geom excluded: " + str(geom) + " for key: " + str(key))
    return resulting_dict_geoms, dict_multi_as_single


def polygonize_reference_data(dict_ref):
    """
    Create a new dictionary with non-overlapping polygons based on a reference data dictionary.

    Args:
        dict_ref (dict): Dictionary of reference geometries.

    Returns:
        dict: Dictionary of non-overlapping polygons.
    """
    arr_ref = []
    for key in dict_ref:
        arr_ref.append((make_valid(dict_ref[key])))
    arr_ref = node(GeometryCollection(arr_ref))
    polygonized = polygonize(arr_ref.geoms)
    dict_ref = {}
    i = 0
    for g in polygonized.geoms:
        if str(g.geom_type) in ["Polygon", "MultiPolygon"]:
            i = i + 1
            key = str(i)  # unique keys will be added (reference_id is lost)
            dict_ref[key] = make_valid(g)
        else:
            logging.warning("geom excluded: " + str(g))
    return dict_ref


def get_breakpoints_zerostreak(x, y):
    """
    Determine the extremes and zero_streaks of a graph based on the derivative.

    Args:
        x (numpy.ndarray): The x values of the graph.
        y (numpy.ndarray): The y values of the graph.

    Returns:
        tuple: A tuple containing:
            - list: List of breakpoints (extremes).
            - list: List of zero_streaks.
    """
    derivative = _numerical_derivative(x, y)
    # plt.plot(x, y, label="y")
    # plt.plot( x, derivative, label="derivative")
    # plt.legend()
    # plt.show()
    extremes = []
    zero_streaks = []
    start_streak = None
    streak = 0
    write_zero_streak = False
    last_extreme = 0
    for i in range(1, len(x)):
        if round(derivative[i], 2) == 0:  # noqa
            streak = streak + 1
            if start_streak is None:
                start_streak = x[i - 1]
        elif streak != 0:
            write_zero_streak = True
        if start_streak is not None and (write_zero_streak or len(x) - 1 == i):
            end_streak = x[i - 1]
            if derivative[i] == 0:
                end_streak = x[i]
            center_streak = start_streak + (end_streak - start_streak) / 2
            zero_streaks.append(
                (start_streak, end_streak, center_streak, streak, last_extreme)
            )
            streak = 0
            start_streak = None
            write_zero_streak = False
            logging.debug("end_streak")
        if (
            i < len(x) - 1
            and round(derivative[i], 2) > 0  # noqa
            and derivative[i - 1] <= derivative[i]
            and derivative[i] >= derivative[i + 1]
        ):
            last_extreme = derivative[i]
            extremes.append((x[i], derivative[i], "maximum"))
        if (
            i < len(x) - 1
            and round(derivative[i], 2) < 0  # noqa
            and derivative[i - 1] >= derivative[i]
            and derivative[i] <= derivative[i + 1]
        ):
            last_extreme = derivative[i]
            extremes.append((x[i], derivative[i], "minimum"))
    for extremum in extremes:
        logging.debug(
            f"breakpoints: relevant_distance:"
            f"{extremum[0]:.2f}, extreme:{extremum[1]:.2f} ({extremum[2]})"
        )
    for st in zero_streaks:
        logging.debug(
            f"zero_streaks: [{st[0]:.2f} - {st[1]:.2f}] - center:{st[2]:.2f}"
            f" - counter:{st[3]:.2f} - min/max-extreme:{st[4]:.2f} "
        )
    # plt.plot(series, afgeleide, label='afgeleide-' + str(key))
    return extremes, zero_streaks


def _numerical_derivative(x, y):
    """
    Calculate the numerical derivative of a graph.

    Parameters:
      x (numpy.ndarray): The x values of the graph.
      y (numpy.ndarray): The y values of the graph.

    Returns:
      numpy.ndarray: The derivative of y with respect to x.
    """

    dx = np.diff(x)
    dy = np.diff(y)
    derivative = dy / dx
    derivative = np.insert(derivative, 0, 0)
    # derivative = np.append(derivative, 0)

    return derivative


def diffs_from_dict_series(
    dict_series: dict[any, dict[float, ProcessResult]],
    dict_thematic: dict[any, BaseGeometry],
    diff_metric: DiffMetric = DiffMetric.CHANGES_AREA,
):
    """
    Calculates a dictionary containing difference metrics for thematic elements based on a distance series.

    Parameters:
    dict_series (dict): A dictionary where keys are thematic IDs and values are dictionaries mapping relative distances to ProcessResult objects.
    dict_thematic (dict): A dictionary where keys are thematic IDs and values are BaseGeometry objects representing the original geometries.
    diff_metric (DiffMetric, optional): The metric to use for calculating differences. Default is DiffMetric.CHANGES_AREA.

    Returns:
    dict: A dictionary where keys are thematic IDs and values are dictionaries mapping relative distances to calculated difference metrics.
    """
    diffs = {}
    # all the relevant distances used to calculate the series
    for thematic_id, results_dict in dict_series.items():
        diffs[thematic_id] = {}

        for rel_dist in results_dict:
            result = results_dict.get(rel_dist, {}).get("result")
            result_diff = results_dict.get(rel_dist, {}).get("result_diff")

            diff = 0
            if (
                result_diff is None
                or result_diff.is_empty
                or result is None
                or result.is_empty
            ):
                diff = 0
            elif diff_metric == DiffMetric.TOTAL_AREA:
                diff = result.area - dict_thematic[thematic_id].area
            elif diff_metric == DiffMetric.TOTAL_PERCENTAGE:
                diff = result.area - dict_thematic[thematic_id].area
                diff = diff * 100 / result.area
            elif diff_metric == DiffMetric.CHANGES_AREA:
                # equals the symmetrical difference, so equal to
                # result_diff_plus.area + result_diff_min.area
                diff = result_diff.area
                # diff = result_diff_plus.area + result_diff_min.area
            elif diff_metric == DiffMetric.CHANGES_PERCENTAGE:
                diff = result_diff.area
                diff = diff * 100 / result.area

            # round, so the detected changes are within 10cm² or 0.1%
            diff = round(diff, 1)
            diffs[thematic_id][rel_dist] = diff

    return diffs


def get_collection(ref_url, limit):
    """
    Fetches a collection of features from a paginated API endpoint.

    This function retrieves a collection of features from a URL that supports
    pagination using a `startIndex` parameter. It iteratively retrieves features in
    chunks of the specified `limit` until no more features are available.

    Args:
        ref_url (str): The base URL of the API endpoint.
        limit (int): The maximum number of features to retrieve per request.

    Returns:
        dict: A dictionary representing the complete GeoJSON feature collection.
            This might be truncated if the total number of features exceeds the
            limitations of the API or server.

    Logs:
        - Debug logs the URL being used for each request during pagination.
    """
    start_index = 0
    collection = {}
    while True:
        url = ref_url + "&startIndex=" + str(start_index)
        logging.debug("called url: " + url)
        json = requests.get(url).json()
        feature_collection = dict(json)
        if (
            "features" not in feature_collection
            or len(feature_collection["features"]) == 0
        ):
            break
        start_index = start_index + limit
        if collection == {}:
            collection = feature_collection
        else:
            collection["features"].extend(feature_collection["features"])  # noqa
        if len(feature_collection["features"]) < limit:
            break
    return collection


def geojson_to_dicts(collection, id_property):
    """
    Converts a GeoJSON collection into dictionaries of geometries and properties.

    Parameters:
    collection (dict): The GeoJSON collection to convert.
    id_property (str): The property name to use as the key for the dictionaries.

    Returns:
    tuple: Two dictionaries:
        - data_dict (dict): A dictionary where keys are the id_property values and values are the geometries.
        - data_dict_properties (dict): A dictionary where keys are the id_property values and values are the properties.
    """
    data_dict = {}
    data_dict_properties = {}
    if collection is None or "features" not in collection:
        return data_dict, data_dict_properties
    for f in collection["features"]:
        key = str(f["properties"][id_property])
        geom = shape(f["geometry"])
        data_dict[key] = make_valid(geom)
        data_dict_properties[key] = f["properties"]
    return data_dict, data_dict_properties


def get_collection_by_partition(
    url, geometry, partition=1000, limit=DOWNLOAD_LIMIT, crs=DEFAULT_CRS
):
    """
    Retrieves a collection of geographic data by partitioning the input geometry.

    Parameters:
    url (str): The base URL for the data source.
    geometry (object): The geometric area to partition and retrieve data for. If None, retrieves data for the entire area.
    partition (int, optional): The number of partitions to divide the geometry into. Default is 1000. If less than 1, no partitioning is done.
    limit (int, optional): The maximum number of items to retrieve. Default is DOWNLOAD_LIMIT.
    crs (str, optional): The coordinate reference system to use. Default is DEFAULT_CRS.

    Returns:
    dict: A collection of geographic data, potentially partitioned by the input geometry.
    """
    collection = {}
    if geometry is None:
        collection = get_collection(
            _add_bbox_to_url(url=url, crs=crs, bbox=None), limit
        )
    elif partition < 1:
        collection = get_collection(
            _add_bbox_to_url(url=url, crs=crs, bbox=get_bbox(geometry)), limit
        )
    else:
        geoms = get_partitions(geometry, partition)
        for g in geoms:
            coll = get_collection(
                _add_bbox_to_url(url=url, crs=crs, bbox=get_bbox(g)), limit
            )
            if collection == {}:
                collection = dict(coll)
            elif "features" in collection and "features" in coll:
                collection["features"].extend(coll["features"])
    return collection


def _add_bbox_to_url(url, crs=DEFAULT_CRS, bbox=None):
    """
    Adds a bounding box (bbox) parameter to the URL for geographic data requests.

    Parameters:
    url (str): The base URL for the data source.
    crs (str, optional): The coordinate reference system to use. Default is DEFAULT_CRS.
    bbox (str, optional): The bounding box coordinates to add to the URL. If None, no bbox is added.

    Returns:
    str: The updated URL with the bbox parameter included, if provided.
    """
    # Load the Base reference data
    if bbox is not None:
        url = url + "&bbox-crs=" + crs + "&bbox=" + bbox
    return url


def merge_process_results(
    result_dict: dict[any, dict[float, ProcessResult]], dict_multi_as_single: dict
) -> dict[any, dict[float, ProcessResult]]:
    """
     Merges processresults in a dictionary from multiple themeIDs into a single themeID.

    Args: result_dict (dict): A dictionary where keys are theme IDs and values are
        process results

    Returns: dict: A new dictionary with merged geometries and remarks (processresults), where keys are global
        theme IDs and values are merged geometries and remarks.

    """
    grouped_results: dict[any, dict[float, ProcessResult]] = {}

    for id_theme, dict_results in result_dict.items():
        if id_theme in dict_multi_as_single.keys():
            id_theme_global = dict_multi_as_single[id_theme]
        else:
            id_theme_global = id_theme
        if id_theme_global not in grouped_results:
            grouped_results[id_theme_global] = dict_results
        else:
            for rel_dist, process_result in dict_results.items():
                for key in process_result:
                    value = process_result[key]  # noqa
                    if isinstance(value, str) and value != "":
                        existing_remark: str = grouped_results[id_theme_global][
                            rel_dist
                        ][
                            key
                        ]  # noqa
                        grouped_results[id_theme_global][rel_dist][key] = (
                            existing_remark + " | " + str(value)
                        )
                        continue
                    elif isinstance(value, BaseGeometry):
                        geom = value
                        if geom.is_empty or geom is None:
                            continue
                        existing: BaseGeometry = grouped_results[id_theme_global][
                            rel_dist
                        ][
                            key
                        ]  # noqa
                        grouped_results[id_theme_global][rel_dist][key] = unary_union(
                            [existing, geom]
                        )  # noqa
    return grouped_results



from brdr.aligner import Aligner
from brdr.enums import GRBType, OpenbaarDomeinStrategy
from brdr.grb import GRBActualLoader
from brdr.oe import OnroerendErfgoedLoader, OEType
from examples import print_brdr_formula
from examples import show_map

if __name__ == "__main__":
    """
    # This example shows the usage of the setting 'multi_as_single_modus'
    # True: (default): All polygons inside a MultiPolygon will be processed seperataly by the algorithm, and merged after processing.
    # False: Multipolygon will be processed directly by the algorithm

    """
    # Example (ErfgoedObject): https://inventaris.onroerenderfgoed.be/erfgoedobjecten/305858
    loader = OnroerendErfgoedLoader([305858], oetype=OEType.EO)
    relevant_distance = 5  # rd is taken very high to show the difference
    od_strategy = OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE
    threshold_circle_ratio = 0.75  # default it is 0.98, but because it are not fully circles in this example we put this on 0.75

    # EXAMPLE of "multi_as_single_modus"=FALSE
    print("EXAMPLE with 'multi_as_single_modus'=False")
    aligner = Aligner(
        multi_as_single_modus=False,
        relevant_distance=relevant_distance,
        od_strategy=od_strategy,
        threshold_circle_ratio=threshold_circle_ratio,
    )
    aligner.load_thematic_data(loader)
    aligner.load_reference_data(
        GRBActualLoader(aligner=aligner, grb_type=GRBType.ADP, partition=1000)
    )
    dict_results = aligner.process()
    aligner.save_results("output/")
    print_brdr_formula(dict_results, aligner)
    show_map(dict_results, aligner.dict_thematic, aligner.dict_reference)

    # WITH "multi_as_single_modus"=True
    print("EXAMPLE with 'multi_as_single_modus'=True")
    aligner = Aligner(
        multi_as_single_modus=True,
        relevant_distance=relevant_distance,
        od_strategy=od_strategy,
        threshold_circle_ratio=threshold_circle_ratio,
    )
    aligner.load_thematic_data(loader)
    aligner.load_reference_data(
        GRBActualLoader(aligner=aligner, grb_type=GRBType.ADP, partition=1000)
    )
    dict_results = aligner.process()
    aligner.save_results("output/")
    print_brdr_formula(dict_results, aligner)
    show_map(dict_results, aligner.dict_thematic, aligner.dict_reference)



import unittest

from brdr.aligner import Aligner
from brdr.oe import OnroerendErfgoedLoader, OEType


class TestOE(unittest.TestCase):
    def test_onroerenderfgoedloader_by_aanduidid(self):
        loader = OnroerendErfgoedLoader(objectids=[120288, 10275], oetype=OEType.AO)
        aligner = Aligner()
        aligner.load_thematic_data(loader)
        assert len(aligner.dict_thematic.keys()) == 2

    def test_onroerenderfgoedloader_by_erfgoedid(self):
        loader = OnroerendErfgoedLoader(objectids=[42549], oetype=OEType.EO)
        aligner = Aligner()
        aligner.load_thematic_data(loader)
        assert len(aligner.dict_thematic.keys()) == 1

    def test_onroerenderfgoedloader_by_bbox(self):
        loader = OnroerendErfgoedLoader(
            bbox=[172000, 172000, 174000, 174000], oetype=OEType.EO
        )
        aligner = Aligner()
        aligner.load_thematic_data(loader)
        assert len(aligner.dict_thematic.keys()) > 0

    def test_onroerenderfgoedloader_by_bbox_and_objectid(self):

        with self.assertRaises(Exception) as context:
            loader = OnroerendErfgoedLoader(
                objectids=[42549],
                bbox=[172000, 172000, 174000, 174000],
                oetype=OEType.EO,
            )

        with self.assertRaises(Exception) as context:
            loader = OnroerendErfgoedLoader(objectids=None, bbox=None, oetype=OEType.EO)



import json
from abc import ABC
from datetime import datetime

import requests as requests
from shapely import make_valid
from shapely.geometry.base import BaseGeometry

from brdr.constants import DATE_FORMAT
from brdr.constants import VERSION_DATE
from brdr.typings import FeatureCollection
from brdr.utils import geojson_to_dicts


class Loader(ABC):
    def __init__(self):
        self.data_dict: dict[any, BaseGeometry] = {}
        self.data_dict_properties: dict[any, dict] = {}
        self.data_dict_source: dict[any, str] = {}
        self.versiondate_info: dict[any, str] = None

    def load_data(self):
        self.data_dict = {x: make_valid(self.data_dict[x]) for x in self.data_dict}
        if self.versiondate_info is not None:
            for key in self.data_dict_properties.keys():
                try:
                    date = datetime.strptime(
                        self.data_dict_properties[key][self.versiondate_info["name"]],
                        self.versiondate_info["format"],
                    )
                except:
                    # Catch, to try extracting only the date with default -date format if specific format does not work
                    date = datetime.strptime(
                        self.data_dict_properties[key][self.versiondate_info["name"]][
                            :10
                        ],
                        DATE_FORMAT,
                    )
                self.data_dict_properties[key][VERSION_DATE] = datetime.strftime(
                    date, DATE_FORMAT
                )

        return self.data_dict, self.data_dict_properties, self.data_dict_source


class DictLoader(Loader):
    def __init__(
        self,
        data_dict: dict[str:BaseGeometry],
        data_dict_properties: dict[str:dict] = {},
        data_dict_source: dict[str:str] = {},
    ):
        super().__init__()
        self.data_dict = data_dict
        self.data_dict_properties = data_dict_properties
        self.data_dict_source = data_dict_source

    def load_data(self):
        return super().load_data()


class GeoJsonLoader(Loader):
    def __init__(self, *, id_property: str = None, _input: FeatureCollection = None):
        super().__init__()
        self.id_property = id_property
        self.input = _input

    def load_data(self):
        self._load_geojson_data()
        return super().load_data()

    def _load_geojson_data(self):
        """
        Load geometries of a GeoJSON and stores them in a dictionary.

        This method processes the thematic data from the input GeoJSON file. It
        iterates through each feature, extracts the relevant properties, converts the
        geometry to a valid shape, and stores it in a dictionary.

        Returns:
            None.
        """
        # THEMATIC PREPARATION
        self.data_dict, self.data_dict_properties = geojson_to_dicts(
            self.input, self.id_property
        )
        return


class GeoJsonFileLoader(GeoJsonLoader):
    def __init__(self, path_to_file, id_property):
        with open(path_to_file, "r") as f:
            _input = json.load(f)
        super().__init__(_input=_input, id_property=id_property)


class GeoJsonUrlLoader(GeoJsonLoader):
    def __init__(self, url, id_property):
        _input = requests.get(url).json()
        super().__init__(_input=_input, id_property=id_property)



from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import GRBActualLoader
from brdr.oe import OnroerendErfgoedLoader, OEType
from examples import print_brdr_formula
from examples import show_map

if __name__ == "__main__":
    # EXAMPLE for a thematic Polygon from Onroerend Erfgoed (https://inventaris.onroerenderfgoed.be/aanduidingsobjecten/131635)

    # Initiate brdr
    aligner = Aligner()
    # Load thematic data from Onroerend Erfgoed
    loader = OnroerendErfgoedLoader(objectids=[131635], oetype=OEType.AO)
    aligner.load_thematic_data(loader)
    # Load reference data: The actual GRB-parcels
    loader = GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=aligner)
    aligner.load_reference_data(loader)

    # PROCESS
    dict_results = aligner.process(relevant_distance=2)

    # GET/SHOW results
    aligner.save_results("output/", formula=True)
    show_map(dict_results, aligner.dict_thematic, aligner.dict_reference)
    print_brdr_formula(dict_results, aligner)



import json
import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import wait
from datetime import datetime
from math import pi
from typing import Iterable

import numpy as np
from shapely import GeometryCollection
from shapely import Polygon
from shapely import STRtree
from shapely import get_parts
from shapely import make_valid
from shapely import remove_repeated_points
from shapely import snap
from shapely import to_geojson
from shapely.geometry.base import BaseGeometry

from brdr import __version__
from brdr.constants import DEFAULT_CRS
from brdr.constants import (
    LAST_VERSION_DATE,
    VERSION_DATE,
    DATE_FORMAT,
    FORMULA_FIELD_NAME,
    EVALUATION_FIELD_NAME,
    FULL_BASE_FIELD_NAME,
    FULL_ACTUAL_FIELD_NAME,
    DIFF_PERCENTAGE_FIELD_NAME,
    DIFF_AREA_FIELD_NAME,
    OD_ALIKE_FIELD_NAME,
    EQUAL_REFERENCE_FEATURES_FIELD_NAME,
)
from brdr.enums import (
    OpenbaarDomeinStrategy,
    Evaluation,
    AlignerResultType,
    AlignerInputType,
)
from brdr.geometry_utils import buffer_neg, safe_unary_union
from brdr.geometry_utils import buffer_neg_pos
from brdr.geometry_utils import buffer_pos
from brdr.geometry_utils import fill_and_remove_gaps
from brdr.geometry_utils import safe_difference
from brdr.geometry_utils import safe_intersection
from brdr.geometry_utils import safe_symmetric_difference
from brdr.geometry_utils import safe_union
from brdr.loader import Loader
from brdr.logger import Logger
from brdr.typings import ProcessResult
from brdr.utils import diffs_from_dict_series, multipolygons_to_singles
from brdr.utils import geojson_from_dict
from brdr.utils import get_breakpoints_zerostreak
from brdr.utils import get_series_geojson_dict
from brdr.utils import merge_process_results
from brdr.utils import write_geojson


###################


class Aligner:
    """
    This class is used to compare and align the thematic data with the reference data.
    The reference data can be loaded in different ways, for example by using the GRB
    data.
    The thematic data can be loaded by using different Loaders: DictLoader, GeojsonLoader,...
    The class can be used to compare and aligne the thematic data with the reference data.
    """

    def __init__(
        self,
        *,
        feedback=None,
        relevant_distance=1,
        relevant_distances=[
            round(k, 1) for k in np.arange(0, 210, 10, dtype=int) / 100
        ],
        threshold_overlap_percentage=50,
        od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE,
        crs=DEFAULT_CRS,
        multi_as_single_modus=True,
        threshold_exclusion_area=0,
        threshold_exclusion_percentage=0,
        buffer_multiplication_factor=1.01,
        threshold_circle_ratio=0.98,
        correction_distance=0.01,
        mitre_limit=10,
        area_limit=None,
        max_workers=None,
    ):
        """
        Initializes the Aligner object

        Args:
            feedback (object, optional): Feedback object that can be added to show
                feedback in QGIS. Defaults to None.
            relevant_distance (int, optional): The relevant distance (in meters) for
                processing. Defaults to 1.
            relevant_distances ([],optional): Relevant distances (in meters) for
                processing
            od_strategy (int, optional): The strategy to determine how to handle
                information outside the reference polygons (Openbaar Domein)
                (default: SNAP_FULL_AREA_ALL_SIDE)
            threshold_overlap_percentage (int, optional): Threshold (%) to determine
                from which overlapping-percentage a reference-polygon has to be included
                when there aren't relevant intersections or relevant differences
                (default 50%).
                When setting this parameter to '-1' the original border for will be returned for cases where nor relevant intersections and relevant differences are found
            crs (str, optional): Coordinate Reference System (CRS) of the data.
                (default EPSG:31370)
            multi_as_single_modus (boolean, optional): Modus to handle multipolygons (Default=True):
                True: input-multipolygons will be split-up into single polygons and handled by the algorithm. After executing the algorithm, the results are merged together.
                False: Multipolygons are directly processed by the algorithm
            threshold_exclusion_percentage (int, optional): Percentage for excluding candidate reference-polygons when overlap(%) is smaller than the threshold(Default=0)
            threshold_exclusion_area (int, optional):Area in m² for excluding candidate reference-polygons when overlap(m²) is smaller than the threshold (Default=0)
            buffer_multiplication_factor (float, optional): Multiplication factor, used to buffer the thematic objects searching for reference borders (buffer= buffer_multiplication_factor*relevant_distance)(Default=1.01)
            threshold_circle_ratio (float, optional): Threshold-value to exclude circles getting processed (perfect circle = 1) based on POLSPY-POPPER algorithm(Default=0.98)
            correction_distance (float, optional): Distance used in a pos_neg_buffer to remove slivers (technical correction) (Default= 0.01 = 1cm )
            mitre_limit (int, optional):buffer-parameter - The mitre ratio is the ratio of the distance from the corner to the end of the mitred offset corner.
                When two line segments meet at a sharp angle, a miter join will extend far beyond the original geometry. (and in the extreme case will be infinitely far.) To prevent unreasonable geometry, the mitre limit allows controlling the maximum length of the join corner.
                Corners with a ratio which exceed the limit will be beveled(Default=10)
            area_limit (int, optional): Maximum area for processing. (default 100000)
            max_workers (int, optional): Amount of workers that is used in ThreadPoolExecutor (for parallel execution) when processing objects for multiple relevant distances. (default None). If set to -1, no parallel exececution is used.

        """
        self.logger = Logger(feedback)
        if relevant_distances is None and relevant_distance is not None:
            relevant_distances = [relevant_distance]
        self.relevant_distances = relevant_distances
        self.od_strategy = od_strategy
        self.threshold_overlap_percentage = threshold_overlap_percentage
        # Area in m² for excluding candidate reference-polygons when overlap(m²) is smaller than the
        # threshold
        self.threshold_exclusion_area = threshold_exclusion_area
        # Percentage for excluding candidate reference-polygons when overlap(%) is smaller than the
        # threshold
        self.threshold_exclusion_percentage = threshold_exclusion_percentage
        self.area_limit = area_limit
        self.max_workers = max_workers
        # Multiplication-factor used in OD-strategy 2 (SNAP-BOTH SIDED) when calculating
        # OD-area to take into account
        self.buffer_multiplication_factor = buffer_multiplication_factor
        # Threshold-value to exclude circles getting processed (perfect circle = 1) based on
        # POLSPY-POPPER algorithm
        self.threshold_circle_ratio = threshold_circle_ratio
        # Distance used in a pos_neg_buffer to remove slivers (technical correction)
        self.correction_distance = correction_distance
        # Buffer parameters:
        # Distance to limit a buffered corner (MITER-join-style parameter)
        #   Explanation and examples:
        #   https://shapely.readthedocs.io/en/stable/reference/shapely.buffer.html
        #   https://postgis.net/docs/ST_Buffer.html
        self.mitre_limit = mitre_limit
        # quad_segments = 8 (by default in shapely)

        # PROCESSING DEFAULTS
        # thematic
        # name of the identifier-field of the thematic data (id has to be unique)
        self.name_thematic_id = "theme_identifier"
        # dictionary to store all thematic geometries to handle
        self.dict_thematic: dict[any, BaseGeometry] = {}
        # dictionary to store properties of the reference-features (optional)
        self.dict_thematic_properties: dict[any, dict] = {}
        # Dict to store source-information of the thematic dictionary
        self.dict_thematic_source: dict[any, str] = {}
        # dictionary to store all unioned thematic geometries
        self.thematic_union = None

        # reference

        # name of the identifier-field of the reference data (id has to be unique,f.e
        # CAPAKEY for GRB-parcels)
        self.name_reference_id = "ref_identifier"
        # dictionary to store all reference geometries
        self.dict_reference: dict[any, BaseGeometry] = {}
        # dictionary to store properties of the reference-features (optional)
        self.dict_reference_properties: dict[any, dict] = {}
        # Dict to store source-information of the reference dictionary
        self.dict_reference_source: dict[any, str] = {}
        # to save a unioned geometry of all reference polygons; needed for calculation
        # in most OD-strategies
        self.reference_union = None

        # results

        # output-dictionaries (all results of process()), grouped by theme_id and relevant_distance
        self.dict_processresults: dict[any, dict[float, ProcessResult]] = {}
        # dictionary with the 'predicted' results, grouped by theme_id and relevant_distance
        self.dict_predictions: dict[any, dict[float, ProcessResult]] = {}
        # dictionary with the 'evaluated predicted' results, grouped by theme_id and relevant_distance
        self.dict_evaluated_predictions: dict[any, dict[float, ProcessResult]] = {}
        # dictionary with the 'evaluated predicted' properties, grouped by theme_id and relevant_distance
        self.dict_evaluated_predictions_properties: dict[any, dict[float, {}]] = {}

        # Coordinate reference system
        # thematic geometries and reference geometries are assumed to be in the same CRS
        # before loading into the Aligner. No CRS-transformation will be performed.
        # When loading data, CRS is expected to be a projected CRS with units in 'meter
        # (m)'.
        # Default EPSG:31370 (Lambert72), alternative: EPSG:3812 (Lambert2008)
        self.CRS = crs
        # this parameter is used to treat multipolygon as single polygons. So polygons
        # with ID splitter are separately evaluated and merged on result.
        self.multi_as_single_modus = multi_as_single_modus
        self.logger.feedback_info("Aligner initialized")

    ##########LOADERS##########################
    ###########################################

    def load_thematic_data(self, loader: Loader):
        """
        Loads the thematic features into the aligner
        :param loader:
        :return:
        """
        self.dict_thematic, self.dict_thematic_properties, self.dict_thematic_source = (
            loader.load_data()
        )
        self.thematic_union = None

    def load_reference_data(self, loader: Loader):
        """
        Loads the reference features into the aligner, and prepares the reference-data for processing
        :param loader:
        :return:
        """
        (
            self.dict_reference,
            self.dict_reference_properties,
            self.dict_reference_source,
        ) = loader.load_data()
        self._prepare_reference_data()

    ##########PROCESSORS#######################
    ###########################################

    def process_geometry(
        self,
        input_geometry: BaseGeometry,
        relevant_distance: float = 1,
        od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE,
        threshold_overlap_percentage=50,
    ) -> ProcessResult:
        """
        method to align a geometry to the reference layer

        Args:
            input_geometry (BaseGeometry): The input geometric object.
            relevant_distance: The relevant distance (in meters) for processing
            od_strategy (int, optional): The strategy to determine how to handle
                information outside the reference polygons (Openbaar Domein)
                (default: SNAP_FULL_AREA_ALL_SIDE)
            threshold_overlap_percentage (int, optional): Threshold (%) to determine
                from which overlapping-percentage a reference-polygon has to be included
                when there aren't relevant intersections or relevant differences
                (default 50%).
                When setting this parameter to '-1' the original border for will be returned for cases where nor relevant intersections and relevant differences are found


        Returns:
            ProcessResult : A dict containing the resulting geometries:

            *   result (BaseGeometry): The resulting output geometry
            *   result_diff (BaseGeometry): The resulting difference output geometry
            *   result_diff_plus (BaseGeometry): The resulting positive difference
                output geometry
            *   result_diff_min (BaseGeometry): The resulting negative difference output
                geometry
            *   relevant_intersection (BaseGeometry): The relevant_intersection
            *   relevant_difference (BaseGeometry): The relevant_difference
            *   remark (str): remarks collected when processing the geoemetry
        """
        if self.area_limit and input_geometry.area > self.area_limit:
            message = "The input geometry is too large to process."
            raise ValueError(message)

        self.logger.feedback_debug("process geometry")
        self.od_strategy = od_strategy
        self.threshold_overlap_percentage = threshold_overlap_percentage
        buffer_distance = relevant_distance / 2
        # combine all parts of the input geometry to one polygon
        input_geometry = safe_unary_union(get_parts(input_geometry))
        input_geometry_inner = buffer_neg(
            input_geometry, relevant_distance
        )  # inner part of the input that must be always available
        if self.od_strategy == OpenbaarDomeinStrategy.EXCLUDE:
            input_geometry_inner = safe_intersection(
                input_geometry_inner, self._get_reference_union()
            )

        # array with all relevant parts of a thematic geometry; initial empty Polygon
        (
            geometry,
            preresult,
            relevant_intersection_array,
            relevant_diff_array,
        ) = self._calculate_intersection_between_geometry_and_od(
            input_geometry, input_geometry_inner, relevant_distance
        )
        # get a list of all ref_ids that are intersecting the thematic geometry
        ref_intersections = self.reference_items.take(
            self.reference_tree.query(geometry)
        ).tolist()
        for key_ref in ref_intersections:
            if key_ref == "24126B0031/00H005":
                pass
            geom_reference = self.dict_reference[key_ref]
            geom_intersection = safe_intersection(geometry, geom_reference)
            if geom_intersection.is_empty or geom_intersection is None:
                continue
            self.logger.feedback_debug("calculate intersection")
            (
                geom,
                relevant_intersection,
                relevant_diff,
            ) = _calculate_geom_by_intersection_and_reference(
                geom_intersection,
                geom_reference,
                input_geometry_inner,
                False,
                buffer_distance,
                self.threshold_overlap_percentage,
                self.threshold_exclusion_percentage,
                self.threshold_exclusion_area,
                self.mitre_limit,
            )
            self.logger.feedback_debug("intersection calculated")
            preresult = self._add_multi_polygons_from_geom_to_array(geom, preresult)
            relevant_intersection_array = self._add_multi_polygons_from_geom_to_array(
                relevant_intersection, relevant_intersection_array
            )
            relevant_diff_array = self._add_multi_polygons_from_geom_to_array(
                relevant_diff, relevant_diff_array
            )
        # UNION INTERMEDIATE LAYERS
        relevant_intersection = safe_unary_union(relevant_intersection_array)
        if relevant_intersection is None or relevant_intersection.is_empty:
            relevant_intersection = Polygon()
        relevant_diff = safe_unary_union(relevant_diff_array)
        if relevant_diff is None or relevant_diff.is_empty:
            relevant_diff = Polygon()

        # POSTPROCESSING
        result_dict = self._postprocess_preresult(
            preresult, geometry, input_geometry_inner, relevant_distance
        )

        result_dict["result_relevant_intersection"] = relevant_intersection
        result_dict["result_relevant_diff"] = relevant_diff

        # make a unary union for each key value in the result dict
        for key in ProcessResult.__annotations__:
            geometry = result_dict.get(key, Polygon())  # noqa
            if isinstance(geometry, BaseGeometry) and not geometry.is_empty:
                geometry = safe_unary_union(geometry)
            result_dict[key] = geometry  # noqa

        return result_dict

    def process(
        self,
        dict_thematic=None,
        relevant_distances: Iterable[float] = None,
        relevant_distance=1,
        od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE,
        threshold_overlap_percentage=50,
    ) -> dict[any, dict[float, ProcessResult]]:
        """
        Calculates the resulting dictionaries for thematic data based on a series of
            relevant distances.

        Args:
            relevant_distances (Iterable[float]): A series of relevant distances
                (in meters) to process
            od_strategy (int, optional): The strategy to determine how to handle
                information outside the reference polygons (Openbaar Domein)
                (default: SNAP_FULL_AREA_ALL_SIDE)
            threshold_overlap_percentage (int, optional): Threshold (%) to determine
                from which overlapping-percentage a reference-polygon has to be included
                when there aren't relevant intersections or relevant differences
                (default 50%).
                When setting this parameter to '-1' the original border for will be returned for cases where nor relevant intersections and relevant differences are found


        Returns:
            dict: A dictionary, for every thematic ID a dictionary with the results for all distances

                {
                    'theme_id_1': {0: (ProcessResult), 0.1:
                        (ProcessResult), ...},
                    'theme_id_2': {0: (ProcessResult), 0.1:
                        (ProcessResult), ...},
                    ...
                }
        """
        if relevant_distances is None:
            relevant_distances = [relevant_distance]
        self.relevant_distances = relevant_distances
        self.od_strategy = od_strategy
        self.threshold_overlap_percentage = threshold_overlap_percentage
        self.logger.feedback_debug("Process series" + str(self.relevant_distances))
        dict_series = {}
        dict_series_queue = {}
        futures = []
        if dict_thematic is None:
            dict_thematic = self.dict_thematic
        dict_multi_as_single = {}
        if self.multi_as_single_modus:
            dict_thematic, dict_multi_as_single = multipolygons_to_singles(
                dict_thematic
            )

        if self.max_workers != -1:
            with ThreadPoolExecutor(
                max_workers=self.max_workers
            ) as executor:  # max_workers=5
                for key, geometry in dict_thematic.items():
                    self.logger.feedback_info(
                        f"thematic id {str(key)} processed with relevant distances (m) [{str(self.relevant_distances)}]"
                    )
                    dict_series[key] = {}
                    dict_series_queue[key] = {}
                    for relevant_distance in self.relevant_distances:
                        try:
                            future = executor.submit(
                                self.process_geometry,
                                geometry,
                                relevant_distance,
                                od_strategy,
                                threshold_overlap_percentage,
                            )
                            futures.append(future)
                            dict_series_queue[key][relevant_distance] = future
                        except ValueError as e:
                            self.logger.feedback_warning(
                                "error for"
                                + f"thematic id {str(key)} processed with relevant distances (m) [{str(self.relevant_distances)}]"
                            )
                            dict_series_queue[key][relevant_distance] = None
                            self.logger.feedback_warning(str(e))
            self.logger.feedback_debug("waiting all started RD calculations")
            wait(futures)
            for id_theme, dict_dist in dict_series_queue.items():
                for relevant_distance, future in dict_dist.items():
                    dict_series[id_theme][relevant_distance] = future.result()
        else:
            for key, geometry in dict_thematic.items():
                self.logger.feedback_info(
                    f"thematic id {str(key)} processed with relevant distances (m) [{str(self.relevant_distances)}]"
                )
                dict_series[key] = {}
                for relevant_distance in self.relevant_distances:
                    try:
                        processed_result = self.process_geometry(
                            geometry,
                            relevant_distance,
                            od_strategy,
                            threshold_overlap_percentage,
                        )
                    except ValueError as e:
                        self.logger.feedback_warning(str(e))
                        processed_result = None

                    dict_series[key][relevant_distance] = processed_result

        if self.multi_as_single_modus:
            dict_series = merge_process_results(dict_series, dict_multi_as_single)

        # Check if geom changes from polygon to multipolygon or vice versa
        for theme_id, dict_dist_results in dict_series.items():
            original_geometry = self.dict_thematic[theme_id]
            original_geometry_length = -1
            if original_geometry.geom_type == "Polygon":
                original_geometry_length = 1
            elif original_geometry.geom_type == "MultiPolygon":
                original_geometry_length = len(original_geometry.geoms)
            for relevant_distance, process_result in dict_dist_results.items():
                resulting_geom = process_result["result"]
                resulting_geometry_length = -1
                if resulting_geom.geom_type == "Polygon":
                    resulting_geometry_length = 1
                elif resulting_geom.geom_type == "MultiPolygon":
                    resulting_geometry_length = len(resulting_geom.geoms)
                if original_geometry_length != resulting_geometry_length:
                    msg = "Difference in amount of polygons"
                    self.logger.feedback_debug(msg)
                    process_result["remark"] = process_result["remark"] + " | " + msg

        self.logger.feedback_info(
            "End of processing series: " + str(self.relevant_distances)
        )
        self.dict_processresults = dict_series

        return self.dict_processresults

    def predictor(
        self,
        dict_thematic=None,
        relevant_distances=[
            round(k, 1) for k in np.arange(0, 310, 10, dtype=int) / 100
        ],
        od_strategy=OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE,
        threshold_overlap_percentage=50,
    ):
        """
        Predicts the 'most interesting' relevant distances for changes in thematic
        elements based on a distance series.

        This function analyzes a set of thematic geometries (`self.dict_thematic`) to
        identify potentially interesting distances where changes occur. It performs
        the following steps:

        1. **Process Distance Series:**
            - Calculates a series of results for different distances specified by
              `relevant_distances`.
            - This calculation might involve functions like `self.process_series`
              (implementation details likely depend on your specific code).

        2. **Calculate Difference Metrics:**
            - Analyzes the results from the distance series to compute difference
              metrics between thematic elements at each distance (using
              `diffs_from_dict_series`).

        3. **Identify Breakpoints and Zero-Streaks:**
            - For each thematic geometry, it identifies potential "breakpoints" where
              the difference metric changes sign (from positive to negative or vice
              versa).
            - It also identifies "zero-streaks" which are consecutive distances with a
              difference metric close to zero (potentially indicating minimal change).

        4. **Predict Interesting Distances:**
            - The function considers distances corresponding to breakpoints and
              zero-streaks as potentially interesting for further analysis.
            - These distances are stored in a dictionary (`dict_predictions`) with the
              thematic element key as the outer key.
            - Additionally, the corresponding results from the distance series for
              those distances are included.

        5. **Filter Results:**
            - The function might further filter the predicted results for each thematic
              element based on the element key (using `filter_resulting_series_by_key`).

        Args:

            relevant_distances (np.ndarray, optional): A series of relevant distances
                (in meters) to process. : A NumPy array of distances to
              be analyzed.
            od_strategy (int, optional): The strategy to determine how to handle
                information outside the reference polygons (Openbaar Domein)
                (default: SNAP_FULL_AREA_ALL_SIDE)
            threshold_overlap_percentage (int, optional): Threshold (%) to determine
                from which overlapping-percentage a reference-polygon has to be included
                when there aren't relevant intersections or relevant differences
                (default 50%).
                When setting this parameter to '-1' the original border for will be returned for cases where nor relevant intersections and relevant differences are found

            relevant_distances (np.ndarray, optional): A NumPy array of distances to
              be analyzed. Defaults to np.arange(0.1, 5.05, 0.1).
            od_strategy (OpenbaarDomeinStrategy, optional): A strategy for handling
              open data in the processing (implementation specific). Defaults to
             OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE.
            threshold_overlap_percentage (int, optional): A percentage threshold for
              considering full overlap in the processing (implementation specific).
             Defaults to 50.

        Returns:
            dict: A dictionary containing predicted interesting distances for each
            thematic element.
                - Keys: Thematic element identifiers from `self.dict_thematic`.
                - Values: Dictionaries with the following structure for each
                   thematic element:
                    - Keys: Distances identified as interesting (breakpoints or
                    zero-streaks).
                    - Values: dicts containing results (likely specific to
                    your implementation) from the distance series for the
                    corresponding distance.
        """

        if dict_thematic is None:
            dict_thematic = self.dict_thematic
        dict_predictions = defaultdict(dict)
        dict_series = self.process(
            relevant_distances=relevant_distances,
            od_strategy=od_strategy,
            threshold_overlap_percentage=threshold_overlap_percentage,
        )

        diffs_dict = diffs_from_dict_series(dict_series, dict_thematic)

        for theme_id, diffs in diffs_dict.items():
            if len(diffs) != len(relevant_distances):
                self.logger.feedback_warning(
                    f"Number of computed diffs for thematic element {theme_id} does "
                    f"not match the number of relevant distances."
                )
                continue
            diff_values = list(diffs.values())
            breakpoints, zero_streaks = get_breakpoints_zerostreak(
                relevant_distances, diff_values
            )
            self.logger.feedback_debug(str(theme_id))
            if len(zero_streaks) == 0:
                self.logger.feedback_debug(
                    "No zero-streaks found for: " + str(theme_id)
                )
            for zs in zero_streaks:
                dict_predictions[theme_id][zs[0]] = dict_series[theme_id][zs[0]]

        # Check if the predicted reldists are unique (and remove duplicated predictions
        dict_predictions_unique = defaultdict(dict)
        for theme_id, dist_results in dict_predictions.items():
            dict_predictions_unique[theme_id] = {}
            predicted_geoms_for_theme_id = []
            for rel_dist, processresults in dist_results.items():
                predicted_geom = processresults["result"]
                if not _equal_geom_in_array(
                    predicted_geom,
                    predicted_geoms_for_theme_id,
                    self.correction_distance,
                    self.mitre_limit,
                ):
                    dict_predictions_unique[theme_id][rel_dist] = processresults
                    predicted_geoms_for_theme_id.append(processresults["result"])
                else:
                    self.logger.feedback_info(
                        f"Duplicate prediction found for key {theme_id} at distance {rel_dist}: Prediction excluded"
                    )

        self.dict_predictions = dict_predictions_unique

        return (
            dict_series,
            self.dict_predictions,
            diffs_dict,
        )

    def evaluate(self, ids_to_evaluate=None, base_formula_field=FORMULA_FIELD_NAME):
        """

        Compares and evaluate input-geometries (with formula). Attributes are added to evaluate and decide if new
        proposals can be used
        affected: list with all IDs to evaluate. all other IDs will be unchanged. If None (default), all self.dict_thematic will be evaluated.
        """
        if ids_to_evaluate is None:
            ids_to_evaluate = list(self.dict_thematic.keys())
        dict_affected = {}
        dict_unaffected = {}
        for id_theme, geom in self.dict_thematic.items():
            if id_theme in ids_to_evaluate:
                dict_affected[id_theme] = geom
            else:
                dict_unaffected[id_theme] = geom
        self.dict_thematic = dict_affected
        # AFFECTED
        dict_series, dict_affected_predictions, diffs = self.predictor(
            dict_thematic=dict_affected, relevant_distances=self.relevant_distances
        )
        dict_predictions_evaluated = {}
        prop_dictionary = {}

        for theme_id in dict_affected.keys():
            dict_predictions_evaluated[theme_id] = {}
            prop_dictionary[theme_id] = {}
            if theme_id not in dict_affected_predictions.keys():
                dist = self.relevant_distances[0]
                prop_dictionary[theme_id][dist] = {}
                props = self._evaluate(
                    id_theme=theme_id,
                    geom_predicted=dict_affected[theme_id],
                    base_formula_field=base_formula_field,
                )
                props[EVALUATION_FIELD_NAME] = Evaluation.TO_CHECK_NO_PREDICTION_5
                dict_predictions_evaluated[theme_id][dist] = dict_series[theme_id][dist]
                prop_dictionary[theme_id][dist] = props
                continue
            dict_predictions_results = dict_affected_predictions[theme_id]
            for dist in sorted(dict_predictions_results.keys()):
                prop_dictionary[theme_id][dist] = {}
                props = self._evaluate(
                    id_theme=theme_id,
                    geom_predicted=dict_predictions_results[dist]["result"],
                    base_formula_field=base_formula_field,
                )
                dict_predictions_evaluated[theme_id][dist] = dict_affected_predictions[
                    theme_id
                ][dist]
                prop_dictionary[theme_id][dist] = props
        # UNAFFECTED
        relevant_distance = 0.0
        # dict_unaffected_series = self.process(dict_thematic=dict_unaffected,relevant_distances=[relevant_distance])
        # for theme_id, dict_unaffected_results in dict_unaffected_series.items():
        for theme_id, geom in dict_unaffected.items():
            dict_predictions_evaluated[theme_id] = {}
            prop_dictionary[theme_id] = {relevant_distance: {}}
            props = self._evaluate(
                id_theme=theme_id,
                geom_predicted=geom,
                base_formula_field=base_formula_field,
            )
            props[EVALUATION_FIELD_NAME] = Evaluation.NO_CHANGE_6
            dict_predictions_evaluated[theme_id][relevant_distance] = {"result": geom}
            prop_dictionary[theme_id][relevant_distance] = props
        self.dict_evaluated_predictions = dict_predictions_evaluated
        self.dict_evaluated_predictions_properties = prop_dictionary
        return dict_predictions_evaluated, prop_dictionary

    def get_brdr_formula(self, geometry: BaseGeometry, with_geom=False):
        """
        Calculates formula-related information based on the input geometry.

        Args:
            geometry (shapely.geometry object): The input geometry.
            with_geom (bool, optional): Whether to include geometry information in the
                output. Defaults to False.

        Returns:
            dict: A dictionary containing formula-related data:

            - "alignment_date": datetime.now().strftime(DATE_FORMAT),
            - "brdr_version": str(__version__),
            - "reference_source": self.dict_reference_source,
            - "full": True if the geometry exists out of all full reference-polygons, else False.
            - "area": Area of the geometry.
            - "reference_features": {
                array of all the reference features the geometry is composed of:
                    -   'full': True if the intersection is the same as the reference
                        geometry, else False.
                    -   'area': Area of the intersection or reference geometry.
                    -   'percentage': Percentage of intersection area relative to the
                        reference geometry.
                    -   'geometry': GeoJSON representation of the intersection (if
                        with_geom is True).},
            - "reference_od": Discription of the OD-part of the geometry (= not covered by reference-features),
        """
        dict_formula = {
            "alignment_date": datetime.now().strftime(DATE_FORMAT),
            "brdr_version": str(__version__),
            "reference_source": self.dict_reference_source,
            "full": True,
            "area": round(geometry.area, 2),
            "reference_features": {},
            "reference_od": None,
        }

        full_total = True
        last_version_date = None

        ref_intersections = self.reference_items.take(
            self.reference_tree.query(geometry)
        ).tolist()
        intersected = []
        for key_ref in ref_intersections:
            geom = None
            version_date = None
            geom_reference = self.dict_reference[key_ref]
            geom_intersection = make_valid(safe_intersection(geometry, geom_reference))
            if geom_intersection.is_empty or geom_intersection is None:
                continue
            intersected.append(geom_intersection)

            perc = round(geom_intersection.area * 100 / geom_reference.area, 2)
            if perc < 0.01:
                continue
            # Add a last_version_date if available in properties
            if (
                key_ref in self.dict_reference_properties
                and VERSION_DATE in self.dict_reference_properties[key_ref]
            ):
                str_version_date = self.dict_reference_properties[key_ref][VERSION_DATE]
                version_date = datetime.strptime(str_version_date, DATE_FORMAT)
                if last_version_date is None and version_date is not None:
                    last_version_date = version_date
                if version_date is not None and version_date > last_version_date:
                    last_version_date = version_date

            if perc > 99.99:
                full = True
                area = round(geom_reference.area, 2)
                perc = 100
                if with_geom:
                    geom = geom_reference
            else:
                full = False
                full_total = False
                area = round(geom_intersection.area, 2)
                if with_geom:
                    geom = geom_intersection

            dict_formula["reference_features"][key_ref] = {
                "full": full,
                "area": area,
                "percentage": perc,
            }
            if version_date is not None:
                dict_formula["reference_features"][key_ref][VERSION_DATE] = (
                    version_date.strftime(DATE_FORMAT)
                )
            if with_geom:
                dict_formula["reference_features"][key_ref]["geometry"] = to_geojson(
                    geom
                )

        dict_formula["full"] = full_total
        if last_version_date is not None:
            dict_formula[LAST_VERSION_DATE] = last_version_date.strftime(DATE_FORMAT)
        geom_od = buffer_pos(
            buffer_neg(
                safe_difference(geometry, safe_unary_union(intersected)),
                self.correction_distance,
                mitre_limit=self.mitre_limit,
            ),
            self.correction_distance,
            mitre_limit=self.mitre_limit,
        )
        if geom_od is not None:
            area_od = round(geom_od.area, 2)
            if area_od > 0:
                dict_formula["reference_od"] = {"area": area_od}
                if with_geom:
                    dict_formula["reference_od"]["geometry"] = to_geojson(geom_od)
        self.logger.feedback_debug(str(dict_formula))
        return dict_formula

    ##########EXPORTERS########################
    ###########################################

    def get_results_as_geojson(
        self,
        resulttype=AlignerResultType.PROCESSRESULTS,
        formula=False,
        attributes=False,
    ):
        """
        get a geojson of a dictionary containing the resulting geometries for all
            'serial' relevant distances. The resulttype can be chosen.
        formula (boolean, Optional): The descriptive formula is added as an attribute to the result
        attributes (boolean, Optional): The original attributes/properties are added to the result
        """
        prop_dictionary = None
        if resulttype == AlignerResultType.PROCESSRESULTS:
            dict_series = self.dict_processresults
        elif resulttype == AlignerResultType.PREDICTIONS:
            dict_series = self.dict_predictions
        elif resulttype == AlignerResultType.EVALUATED_PREDICTIONS:
            dict_series = self.dict_evaluated_predictions
            prop_dictionary = self.dict_evaluated_predictions_properties
        else:
            raise (ValueError, "AlignerResultType unknown")
        if dict_series is None or dict_series == {}:
            self.logger.feedback_warning(
                "Empty results: No calculated results to export."
            )
            return {}
        if prop_dictionary is None:
            prop_dictionary = defaultdict(dict)

        for theme_id, results_dict in dict_series.items():
            for relevant_distance, process_results in results_dict.items():
                if relevant_distance not in prop_dictionary[theme_id]:
                    prop_dictionary[theme_id][relevant_distance] = {}
                if attributes:
                    for attr, value in self.dict_thematic_properties[theme_id].items():
                        prop_dictionary[theme_id][relevant_distance][attr] = value
                if (
                    formula
                ):  # and not (theme_id in prop_dictionary and relevant_distance in prop_dictionary[theme_id] and NEW_FORMULA_FIELD_NAME in prop_dictionary[theme_id][relevant_distance]):
                    result = process_results["result"]
                    formula = self.get_brdr_formula(result)
                    prop_dictionary[theme_id][relevant_distance][FORMULA_FIELD_NAME] = (
                        json.dumps(formula)
                    )
        return get_series_geojson_dict(
            dict_series,
            crs=self.CRS,
            id_field=self.name_thematic_id,
            series_prop_dict=prop_dictionary,
        )

    def get_input_as_geojson(self, inputtype=AlignerInputType.REFERENCE):
        """
        get a geojson of the input polygons (thematic or reference-polygons)
        """

        if inputtype == AlignerInputType.THEMATIC:
            dict_to_geojson = self.dict_thematic
            dict_properties = self.dict_thematic_properties
            property_id = self.name_thematic_id
        elif inputtype == AlignerInputType.REFERENCE:
            dict_to_geojson = self.dict_reference
            dict_properties = self.dict_reference_properties
            property_id = self.name_reference_id
        else:
            raise (ValueError, "AlignerInputType unknown")
        if dict_to_geojson is None or dict_to_geojson == {}:
            self.logger.feedback_warning("Empty input: No input to export.")
            return {}
        return geojson_from_dict(
            dict_to_geojson,
            self.CRS,
            property_id,
            prop_dict=dict_properties,
            geom_attributes=False,
        )

    def save_results(
        self, path, resulttype=AlignerResultType.PROCESSRESULTS, formula=True
    ):
        """
        Exports analysis results (as geojson) to path.

        This function exports 6 GeoJSON files containing the analysis results to the
        specified `path`.

        Args:
        path (str): The path to the directory where the GeoJSON files will be saved.
        formula (bool, optional): Whether to include formula-related information
            in the output. Defaults to True.

        Details of exported files:
        - result.geojson: Contains the original thematic data from `
          self.dict_result`.
        - result_diff.geojson: Contains the difference between the original
          and predicted data from `self.dict_result_diff`.
        - result_diff_plus.geojson: Contains results for areas that are
          added (increased area).
        - result_diff_min.geojson: Contains results for areas that are
          removed (decreased area).
        - result_relevant_intersection.geojson: Contains the areas with
          relevant intersection that has to be included in the result.
        - result_relevant_difference.geojson: Contains the areas with
          relevant difference that has to be excluded from the result.
        """

        fcs = self.get_results_as_geojson(
            formula=formula,
            resulttype=resulttype,
        )
        for name, fc in fcs.items():
            write_geojson(
                os.path.join(path, resulttype.value + "_" + name + ".geojson"), fc
            )

    def get_thematic_union(self):
        """
        returns a unary_unioned geometry from all the thematic geometries
        :return:
        """
        if self.thematic_union is None:
            self.thematic_union = safe_unary_union(list(self.dict_thematic.values()))
        return self.thematic_union

    def _prepare_reference_data(self):
        """
        Prepares reference data for spatial queries and analysis.
        It performs the following tasks:
        1. **Optimizes spatial queries:**
            - Creates a Spatial Relationship Tree (STRtree) using `STRtree` for
              efficient spatial queries against the reference data in
              `self.dict_reference`.
            - Converts the dictionary keys (reference identifiers) to a NumPy array
              for potential performance benefits in future operations.

        2. **Clears reference union:**
            - Sets `self.reference_union` to `None`. This variable stores the combined
              geometry of all reference data, and it's cleared here to indicate that
              it needs to be recalculated if requested later.

        Returns:
            None
        """
        # create an SRTree for performance optimisation
        self.logger.feedback_info(
            "length of reference_dict: " + str(len(self.dict_reference))
        )
        self.reference_tree = STRtree(list(self.dict_reference.values()))
        self.reference_items = np.array(list(self.dict_reference.keys()))
        # clear the reference_union, so it will be recalculated on request when needed
        self.reference_union = None
        return

    def _calculate_intersection_between_geometry_and_od(
        self, geometry, input_geometry_inner, relevant_distance
    ):
        """
        Calculates the intersecting parts between a thematic geometry and the openbaardomein( domain, not coverd by reference-polygons)
        :param geometry:
        :param relevant_distance:
        :return:
        """
        # Calculate the intersection between thematic and Openbaar Domein
        buffer_distance = relevant_distance / 2
        relevant_intersection_array = []
        relevant_difference_array = []
        geom_thematic_od = Polygon()

        if self.od_strategy == OpenbaarDomeinStrategy.EXCLUDE:
            # Completely exclude everything that is not on the reference layer
            self.logger.feedback_debug("OD-strategy EXCLUDE")
            # Remove from the thematic layer all parts that are not on the reference
            # layer
            # !!this strategy adapts the input-geometry!!
            geometry = safe_intersection(geometry, self._get_reference_union())
        elif self.od_strategy == OpenbaarDomeinStrategy.AS_IS:
            # All parts that are not covered by the reference layer are added to the
            #         resulting geometry AS IS
            self.logger.feedback_debug("OD-strategy AS IS")
            # all OD-parts wil be added AS IS
            geom_thematic_od = safe_difference(geometry, self._get_reference_union())
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE:
            # Everything that falls within the relevant distance over the
            #  plot boundary is snapped to the plot.
            #  Only the inner-reference-boundaries are used.
            #  The outer-reference-boundaries are not used.
            self.logger.feedback_debug("OD-strategy SNAP_SINGLE_SIDE")
            # geom of OD
            geom_od = safe_difference(geometry, self._get_reference_union())
            # only the relevant parts of OD
            geom_od_neg_pos = buffer_neg_pos(
                geom_od,
                buffer_distance,
                mitre_limit=self.mitre_limit,
            )
            # geom_thematic_od = safe_intersection(geom_od_neg_pos,geom_od)# resulting
            # thematic OD
            geom_od_neg_pos_buffered = buffer_pos(
                geom_od_neg_pos,
                buffer_distance,
                mitre_limit=self.mitre_limit,
            )  # include parts
            geom_thematic_od = safe_intersection(
                geom_od_neg_pos_buffered, geom_od
            )  # resulting thematic OD
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_ALL_SIDE:
            # Everything that falls within the relevant distance over
            # the plot boundary is snapped to the plot.
            #  Inner-reference-boundaries and outer-reference-boundaries are used.
            self.logger.feedback_debug("OD-strategy SNAP BOTH SIDED")
            (
                geom_thematic_od,
                relevant_difference_array,
                relevant_intersection_array,
            ) = self._od_snap_all_side(
                geometry, input_geometry_inner, relevant_distance
            )
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_FULL_AREA_SINGLE_SIDE:
            # Strategy useful for bigger areas.
            # integrates the entire inner area of the input geometry,
            # so Openbaar Domein of the inner area is included in the result
            # Combines SNAP_SINGLE_SIDE with the inner area
            self.logger.feedback_debug(
                "OD-strategy Full-area-variant of OD-SNAP_SINGLE_SIDE"
            )
            geom_thematic_od = self._od_full_area(geometry, relevant_distance)
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_FULL_AREA_ALL_SIDE:
            # Strategy useful for bigger areas.
            # integrates the entire inner area of the input geometry,
            # so Openbaar Domein of the inner area is included in the result
            # Combines SNAP_ALL_SIDE with the inner area
            self.logger.feedback_debug(
                "OD-strategy Full-area-variant of OD-SNAP_ALL_SIDE"
            )
            # first part is a copy of OD_ALL_SIDE
            (
                geom_thematic_od,
                relevant_difference_array,
                relevant_intersection_array,
            ) = self._od_snap_all_side(
                geometry, input_geometry_inner, relevant_distance
            )
            # This part is a copy of SNAP_FULL_AREA_SINGLE_SIDE
            geom_theme_od_min_clipped_plus_buffered_clipped = self._od_full_area(
                geometry, relevant_distance
            )
            # UNION the calculation of  OD-SNAP_ALL_SIDE with FULL AREA of
            # OD-SNAP_FULL_AREA_SINGLE_SIDE
            geom_thematic_od = safe_union(
                geom_theme_od_min_clipped_plus_buffered_clipped, geom_thematic_od
            )

        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE_VARIANT_1:
            # OD-strategy SNAP_SINGLE_SIDE - variant 1:
            # Everything that falls within the relevant distance over the
            #  plot boundary is snapped to the plot.
            #  Only the inner-reference-boundaries are used.
            #  The outer-reference-boundaries are not used.
            self.logger.feedback_debug("OD-strategy SNAP_SINGLE_SIDE - variant 1")
            # geom of OD
            geom_od = safe_difference(geometry, self._get_reference_union())
            # only the relevant parts of OD
            geom_od_neg_pos = buffer_neg_pos(
                geom_od,
                buffer_distance,
                mitre_limit=self.mitre_limit,
            )
            # resulting thematic OD
            geom_thematic_od = safe_intersection(geom_od_neg_pos, geom_od)
        elif self.od_strategy == OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE_VARIANT_2:
            # OD-strategy SNAP_SINGLE_SIDE - variant 2:
            # Everything that falls within the relevant distance over the
            #  plot boundary is snapped to the plot.
            #  Only the inner-reference-boundaries are used.
            #  The outer-reference-boundaries are not used.
            self.logger.feedback_debug("OD-strategy SNAP_SINGLE_SIDE - variant 2")
            geom_thematic_od = Polygon()
            pass

        # ADD THEMATIC_OD
        preresult = self._add_multi_polygons_from_geom_to_array(geom_thematic_od, [])
        return (
            geometry,
            preresult,
            relevant_intersection_array,
            relevant_difference_array,
        )

    def _od_full_area(self, geometry, relevant_distance):
        buffer_distance = relevant_distance / 2
        geom_theme_od = safe_difference(geometry, self._get_reference_union())
        geom_theme_min_buffered = buffer_neg(
            buffer_pos(
                buffer_neg(
                    geometry,
                    relevant_distance,
                    mitre_limit=self.mitre_limit,
                ),
                buffer_distance,
                mitre_limit=self.mitre_limit,
            ),
            buffer_distance,
            mitre_limit=self.mitre_limit,
        )
        geom_theme_od_clipped_min_buffered = safe_intersection(
            geom_theme_min_buffered, geom_theme_od
        )
        geom_theme_od_min_clipped_plus_buffered = buffer_pos(
            geom_theme_od_clipped_min_buffered,
            relevant_distance,
            mitre_limit=self.mitre_limit,
        )
        geom_theme_od_min_clipped_plus_buffered_clipped = safe_intersection(
            geom_theme_od_min_clipped_plus_buffered, geom_theme_od
        )
        geom_thematic_od = geom_theme_od_min_clipped_plus_buffered_clipped
        return geom_thematic_od

    def _od_snap_all_side(self, geometry, input_geometry_inner, relevant_distance):
        buffer_distance = relevant_distance / 2
        relevant_difference_array = []
        relevant_intersection_array = []
        geom_thematic_buffered = make_valid(
            buffer_pos(
                geometry,
                self.buffer_multiplication_factor * relevant_distance,
                mitre_limit=self.mitre_limit,
            )
        )
        clip_ref_thematic_buffered = safe_intersection(
            self._get_reference_union(), geom_thematic_buffered
        )
        geom_reference = safe_difference(
            geom_thematic_buffered, clip_ref_thematic_buffered
        )  # Both OD-parts are SNAPPED added
        geom_thematic = geometry
        if geom_reference.is_empty or geom_reference is None:
            geom_thematic_od = geom_reference
        else:
            geom_intersection = safe_intersection(geom_reference, geom_thematic)
            (
                geom_thematic_od,
                geom_relevant_intersection,
                geom_relevant_diff,
            ) = _calculate_geom_by_intersection_and_reference(
                geom_intersection,
                geom_reference,
                input_geometry_inner,
                True,
                buffer_distance,
                self.threshold_overlap_percentage,
                self.threshold_exclusion_percentage,
                self.threshold_exclusion_area,
                self.mitre_limit,
            )
            relevant_intersection_array = self._add_multi_polygons_from_geom_to_array(
                geom_relevant_intersection, []
            )
            relevant_difference_array = self._add_multi_polygons_from_geom_to_array(
                geom_relevant_diff, []
            )
        return geom_thematic_od, relevant_difference_array, relevant_intersection_array

    def _get_reference_union(self):
        """
        returns a unary_unioned geometry from all the referene geometries
        :return:
        """
        if self.reference_union is None:
            self.reference_union = safe_unary_union(list(self.dict_reference.values()))
        return self.reference_union

    def _postprocess_preresult(
        self, preresult, geom_thematic, input_geometry_inner, relevant_distance
    ) -> ProcessResult:
        """
        Postprocess the preresult with the following actions to create the final result
        *Corrections for areas that differ more than the relevant distance
        *slivers
        *Inner holes (donuts) /multipolygons
        *validity
        *Circles (Polspy-popper)
        *Null/Empty-values

        Args:
            preresult (list): An existing list with all the elements of the preresult
            geom_thematic (BaseGeometry): The input geometry

        Returns:
            ProcessResult: A dictionary containing the resulting output geometries:

            *   result (BaseGeometry): The resulting output geometry
            *   result_diff (BaseGeometry): The resulting difference output geometry
            *   result_diff_plus (BaseGeometry): The resulting positive difference
                output geometry
            *   result_diff_min (BaseGeometry): The resulting negative difference output
                geometry
            *   remark (str): Remark when processing the geometry
        """
        # Process array
        remark = ""
        buffer_distance = relevant_distance / 2
        result = []
        geom_thematic = make_valid(geom_thematic)
        preresult.append(input_geometry_inner)
        geom_preresult = safe_unary_union(preresult)

        if not (geom_thematic is None or geom_thematic.is_empty):
            # Correction for circles
            # calculate ratio to see if it is a circle, and keep the original geometry
            #  if a circle: (Polsby-popper score)
            if (
                4 * pi * (geom_thematic.area / (geom_thematic.length**2))
                > self.threshold_circle_ratio
            ):
                remark = "Circle detected: -->resulting geometry = original geometry"
                self.logger.feedback_debug(remark)
                return {"result": geom_thematic, "remark": remark}

            # Correction for unchanged geometries
            if safe_symmetric_difference(geom_preresult, geom_thematic).is_empty:
                remark = "Unchanged geometry: -->resulting geometry = original geometry"
                self.logger.feedback_debug(remark)
                return {"result": geom_thematic, "remark": remark}

        # Corrections for areas that differ more than the relevant distance
        geom_thematic_dissolved = buffer_pos(
            buffer_neg(
                buffer_pos(
                    geom_preresult,
                    self.correction_distance,
                    mitre_limit=self.mitre_limit,
                ),
                2 * self.correction_distance,
                mitre_limit=self.mitre_limit,
            ),
            self.correction_distance,
            mitre_limit=self.mitre_limit,
        )
        # geom_symdiff = self._safe_symmetric_difference(geom_thematic,
        # geom_thematic_dissolved)
        geom_diff_add = safe_difference(geom_thematic, geom_thematic_dissolved)
        geom_diff_delete = safe_difference(geom_thematic_dissolved, geom_thematic)
        geom_diff_removed = safe_difference(
            geom_thematic_dissolved,
            safe_intersection(
                geom_diff_delete,
                buffer_neg_pos(
                    geom_diff_delete,
                    buffer_distance,
                    mitre_limit=self.mitre_limit,
                ),
            ),
        )
        geom_diff_removed_added = safe_union(
            geom_diff_removed,
            safe_intersection(
                geom_diff_add,
                buffer_neg_pos(
                    geom_diff_add,
                    buffer_distance,
                    mitre_limit=self.mitre_limit,
                ),
            ),
        )
        geom_thematic_preresult = buffer_pos(
            buffer_neg(
                buffer_pos(
                    geom_diff_removed_added,
                    self.correction_distance,
                    mitre_limit=self.mitre_limit,
                ),
                2 * self.correction_distance,
                mitre_limit=self.mitre_limit,
            ),
            self.correction_distance,
            mitre_limit=self.mitre_limit,
        )
        # Correction for Inner holes(donuts) / multipolygons
        # fill and remove gaps
        geom_thematic_cleaned_holes = fill_and_remove_gaps(
            geom_thematic_preresult, buffer_distance
        )
        geom_thematic_result = buffer_pos(
            buffer_neg(
                buffer_pos(
                    geom_thematic_cleaned_holes,
                    self.correction_distance,
                    mitre_limit=self.mitre_limit,
                ),
                2 * self.correction_distance,
                mitre_limit=self.mitre_limit,
            ),
            self.correction_distance,
            mitre_limit=self.mitre_limit,
        )
        geom_thematic_result = make_valid(remove_repeated_points(geom_thematic_result))

        # Correction for empty preresults
        if geom_thematic_result.is_empty or geom_thematic_result is None:
            remark = "Calculated empty result: -->original geometry returned"
            self.logger.feedback_warning(remark)

            geom_thematic_result = geom_thematic
            # geom_thematic_result = Polygon() #If we return an empty geometry, the feature disappears, so we return the original geometry

        # group all initial multipolygons into a new resulting dictionary
        result.append(geom_thematic_result)

        # create all resulting geometries
        geom_thematic_result = safe_unary_union(result)

        # negative and positive buffer is added to the difference-calculations, to
        # remove 'very small' differences (smaller than the correction distance)
        geom_result_diff = buffer_pos(
            buffer_neg(
                safe_symmetric_difference(
                    geom_thematic_result,
                    geom_thematic,
                ),
                self.correction_distance,
                mitre_limit=self.mitre_limit,
            ),
            self.correction_distance,
            mitre_limit=self.mitre_limit,
        )
        geom_result_diff_plus = buffer_pos(
            buffer_neg(
                safe_difference(
                    geom_thematic_result,
                    geom_thematic,
                ),
                self.correction_distance,
                mitre_limit=self.mitre_limit,
            ),
            self.correction_distance,
            mitre_limit=self.mitre_limit,
        )
        geom_result_diff_min = buffer_pos(
            buffer_neg(
                safe_difference(
                    geom_thematic,
                    geom_thematic_result,
                ),
                self.correction_distance,
                mitre_limit=self.mitre_limit,
            ),
            self.correction_distance,
            mitre_limit=self.mitre_limit,
        )
        # geom_result_diff_plus = safe_difference(geom_thematic_result, geom_thematic)
        # geom_result_diff_min = safe_difference(geom_thematic, geom_thematic_result)

        return {
            "result": geom_thematic_result,
            "result_diff": geom_result_diff,
            "result_diff_plus": geom_result_diff_plus,
            "result_diff_min": geom_result_diff_min,
            "remark": remark,
        }

    def _evaluate(
        self, id_theme, geom_predicted, base_formula_field=FORMULA_FIELD_NAME
    ):
        """
        function that evaluates a predicted geometry and returns a properties-dictionary
        """
        threshold_od_percentage = 1
        properties = {
            FORMULA_FIELD_NAME: "",
            EVALUATION_FIELD_NAME: Evaluation.TO_CHECK_PREDICTION_4,
            FULL_BASE_FIELD_NAME: None,
            FULL_ACTUAL_FIELD_NAME: None,
            OD_ALIKE_FIELD_NAME: None,
            EQUAL_REFERENCE_FEATURES_FIELD_NAME: None,
            DIFF_PERCENTAGE_FIELD_NAME: None,
            DIFF_AREA_FIELD_NAME: None,
        }
        actual_formula = self.get_brdr_formula(geom_predicted)
        properties[FORMULA_FIELD_NAME] = json.dumps(actual_formula)
        base_formula = None
        if (
            id_theme in self.dict_thematic_properties
            and base_formula_field in self.dict_thematic_properties[id_theme]
        ):
            base_formula = json.loads(
                self.dict_thematic_properties[id_theme][base_formula_field]
            )

        if base_formula is None or actual_formula is None:
            properties[EVALUATION_FIELD_NAME] = Evaluation.TO_CHECK_NO_PREDICTION_5
            return properties
        properties[FULL_BASE_FIELD_NAME] = base_formula["full"]
        properties[FULL_ACTUAL_FIELD_NAME] = actual_formula["full"]
        od_alike = False
        if (
            base_formula["reference_od"] is None
            and actual_formula["reference_od"] is None
        ):
            od_alike = True
        elif (
            base_formula["reference_od"] is None
            or actual_formula["reference_od"] is None
        ):
            od_alike = False
        elif (
            abs(
                base_formula["reference_od"]["area"]
                - actual_formula["reference_od"]["area"]
            )
            * 100
            / base_formula["reference_od"]["area"]
        ) < threshold_od_percentage:
            od_alike = True
        properties[OD_ALIKE_FIELD_NAME] = od_alike

        equal_reference_features = True
        if (
            base_formula["reference_features"].keys()
            == actual_formula["reference_features"].keys()
        ):
            equal_reference_features = True
            max_diff_area_reference_feature = 0
            max_diff_percentage_reference_feature = 0
            for key in base_formula["reference_features"].keys():
                if (
                    base_formula["reference_features"][key]["full"]
                    != actual_formula["reference_features"][key]["full"]
                ):
                    equal_reference_features = False

                diff_area_reference_feature = abs(
                    base_formula["reference_features"][key]["area"]
                    - actual_formula["reference_features"][key]["area"]
                )
                diff_percentage_reference_feature = (
                    abs(
                        base_formula["reference_features"][key]["area"]
                        - actual_formula["reference_features"][key]["area"]
                    )
                    * 100
                    / base_formula["reference_features"][key]["area"]
                )
                if diff_area_reference_feature > max_diff_area_reference_feature:
                    max_diff_area_reference_feature = diff_area_reference_feature
                if (
                    diff_percentage_reference_feature
                    > max_diff_percentage_reference_feature
                ):
                    max_diff_percentage_reference_feature = (
                        diff_percentage_reference_feature
                    )
            properties[EQUAL_REFERENCE_FEATURES_FIELD_NAME] = equal_reference_features
            properties[DIFF_AREA_FIELD_NAME] = max_diff_area_reference_feature
            properties[DIFF_PERCENTAGE_FIELD_NAME] = (
                max_diff_percentage_reference_feature
            )
        # EVALUATION
        if (
            equal_reference_features
            and od_alike
            and base_formula["full"]
            and actual_formula["full"]
        ):
            properties[EVALUATION_FIELD_NAME] = Evaluation.EQUALITY_EQUAL_FORMULA_FULL_1
        elif (
            equal_reference_features
            and od_alike
            and base_formula["full"] == actual_formula["full"]
        ):
            properties[EVALUATION_FIELD_NAME] = Evaluation.EQUALITY_EQUAL_FORMULA_2
        elif base_formula["full"] and actual_formula["full"] and od_alike:
            properties[EVALUATION_FIELD_NAME] = Evaluation.EQUALITY_FULL_3
        # elif base_formula["full"] == actual_formula["full"] and od_alike:#TODO evaluate when not-full-parcels?
        #    properties[EVALUATION_FIELD_NAME] = Evaluation.EQUALITY_NON_FULL
        else:
            properties[EVALUATION_FIELD_NAME] = Evaluation.TO_CHECK_PREDICTION_4
        return properties

    @staticmethod
    def _add_multi_polygons_from_geom_to_array(geom: BaseGeometry, array):
        """
        Append valid polygons and multipolygons extracted from a given geometry to an
        existing array.

        Args:
            geom (BaseGeometry): The input geometry to process.
            array (list): An existing list to store valid polygons and multipolygons.

        Returns:
            list: A list containing valid polygons and multipolygons extracted from the
                input geometry.
        """
        if geom.is_empty or geom is None:
            # If the input geometry is empty or None, do nothing.
            pass
        else:
            # Create a GeometryCollection from the input geometry.
            geometry_collection = GeometryCollection(geom)  # noqa
            for g in geometry_collection.geoms:
                # Ensure each sub-geometry is valid.
                g = make_valid(g)
                if str(g.geom_type) in ["Polygon", "MultiPolygon"]:
                    # Append valid polygons and multipolygons to the array.
                    array.append(g)
        return array


def _calculate_geom_by_intersection_and_reference(
    geom_intersection: BaseGeometry,
    geom_reference: BaseGeometry,
    input_geometry_inner: BaseGeometry,
    is_openbaar_domein,
    buffer_distance,
    threshold_overlap_percentage,
    threshold_exclusion_percentage,
    threshold_exclusion_area,
    mitre_limit,
):
    """
    Calculates the geometry based on intersection and reference geometries.

    Args:
        geom_intersection (BaseGeometry): The intersection geometry.
        geom_reference (BaseGeometry): The reference geometry.
        is_openbaar_domein (bool): A flag indicating whether it's a public domain
            (area not covered with reference polygon).
        threshold_exclusion_percentage (int): The threshold exclusion percentage.
        threshold_exclusion_area (int): The threshold exclusion area.
        buffer_distance (float): The buffer distance.
        threshold_overlap_percentage (int): The threshold overlap percentage.

    Returns:
        tuple: A tuple containing the resulting geometries:

        *   geom: BaseGeometry or None: The resulting geometry or None if conditions
            are not met.
        *   geom_relevant_intersection: BaseGeometry or None: The relevant
            intersection.
        *   geom_relevant_difference: BaseGeometry or None: The relevant difference.

    Notes:
        -   If the reference geometry area is 0, the overlap is set to 100%.
        -   If the overlap is less than relevant_OVERLAP_PERCENTAGE or the
            intersection area is less than relevant_OVERLAP_AREA, None is returned.
        -   Otherwise, the relevant intersection and difference geometries are
            calculated.
        -   If both relevant intersection and difference are non-empty, the final
            geometry is obtained by applying safe intersection and buffering.
        -   If only relevant intersection is non-empty, the result is the reference
            geometry.
        -   If only relevant difference is non-empty, the result is None.
    """

    if geom_reference.area == 0:
        overlap = 100

    else:
        overlap = geom_intersection.area * 100 / geom_reference.area

    if (
        overlap < threshold_exclusion_percentage
        or geom_intersection.area < threshold_exclusion_area
    ):
        return Polygon(), Polygon(), Polygon()

    geom_difference = safe_difference(geom_reference, geom_intersection)
    geom_relevant_intersection = buffer_neg(
        geom_intersection,
        buffer_distance,
        mitre_limit=mitre_limit,
    )
    geom_intersection_inner = safe_intersection(geom_intersection, input_geometry_inner)

    geom_relevant_difference = buffer_neg(
        geom_difference,
        buffer_distance,
        mitre_limit=mitre_limit,
    )
    if (
        not geom_intersection_inner.is_empty
        and geom_relevant_intersection.is_empty
        and not geom_relevant_difference.is_empty
    ):
        geom_x = safe_intersection(
            geom_intersection, buffer_pos(geom_intersection_inner, 1 * buffer_distance)
        )
        geom_x = snap(geom_x, geom_reference, 1 * buffer_distance)
        geom = safe_intersection(geom_intersection, geom_x)
    elif (
        not geom_relevant_intersection.is_empty
        and not geom_relevant_difference.is_empty
    ):
        # relevant intersection and relevant difference

        # calculate part where difference is removed
        geom_x = safe_intersection(
            geom_reference,
            safe_difference(
                geom_reference,
                safe_intersection(
                    geom_difference,
                    buffer_neg_pos(
                        geom_difference,
                        buffer_distance,
                        mitre_limit=mitre_limit,
                    ),
                ),
            ),
        )
        # calculate part that is relevant
        geom_y = buffer_pos(
            buffer_neg_pos(
                geom_x,
                buffer_distance,
                mitre_limit=mitre_limit,
            ),
            buffer_distance,
            mitre_limit=mitre_limit,
        )
        geom_y = safe_unary_union(
            [geom_y, geom_intersection_inner]
        )  # add inner part that has to be present
        geom = safe_intersection(
            geom_x,
            geom_y,
        )
        # when calculating for OD, we create a 'virtual parcel'. When calculating this
        # virtual parcel, it is buffered to take outer boundaries into account.
        # This results in a side effect that there are extra non-logical parts included
        # in the result. The function below tries to exclude these non-logical parts.
        # see eo_id 206363 with relevant distance=0.2m and SNAP_ALL_SIDE
        if is_openbaar_domein:
            geom = _get_relevant_polygons_from_geom(geom, buffer_distance, mitre_limit)
    elif not geom_relevant_intersection.is_empty and geom_relevant_difference.is_empty:
        geom = geom_reference
    elif geom_relevant_intersection.is_empty and not geom_relevant_difference.is_empty:
        geom = geom_relevant_intersection  # (=empty geometry)
    else:
        if is_openbaar_domein:
            geom = geom_relevant_intersection  # (=empty geometry)
        # geom = snap_geom_to_reference (geom_intersection, geom_reference,
        # relevant_distance)
        elif threshold_overlap_percentage < 0:
            # if we take a value of -1, the original border will be used
            geom = geom_intersection
        elif overlap > threshold_overlap_percentage:
            geom = geom_reference
        else:
            geom = geom_relevant_intersection  # (=empty geometry)
    return geom, geom_relevant_intersection, geom_relevant_difference


def _get_relevant_polygons_from_geom(
    geometry: BaseGeometry, buffer_distance: float, mitre_limit
):
    """
    Get only the relevant parts (polygon) from a geometry.
    Points, Lines and Polygons smaller than relevant distance are excluded from the
    result
    """
    if not geometry or geometry.is_empty:
        # If the input geometry is empty or None, do nothing.
        return geometry
    else:
        geometry = safe_unary_union(geometry)
        # Create a GeometryCollection from the input geometry.
        geometry_collection = GeometryCollection(geometry)
        array = []
        for g in geometry_collection.geoms:
            # Ensure each sub-geometry is valid.
            g = make_valid(g)
            if str(g.geom_type) in ["Polygon", "MultiPolygon"]:
                relevant_geom = buffer_neg(
                    g,
                    buffer_distance,
                    mitre_limit=mitre_limit,
                )
                if relevant_geom is not None and not relevant_geom.is_empty:
                    array.append(g)
    return safe_unary_union(array)


def _equal_geom_in_array(geom, geom_array, correction_distance, mitre_limit):
    """
    Check if a predicted geometry is equal to other predicted geometries in a list.
    Equality is defined as there is the symmetrical difference is smaller than the CORRECTION DISTANCE
    Returns True if one of the elements is equal, otherwise False
    """
    for g in geom_array:
        # if safe_equals(geom,g):
        if buffer_neg(
            safe_symmetric_difference(geom, g),
            correction_distance,
            mitre_limit=mitre_limit,
        ).is_empty:
            return True
    return False


def _check_equality(
    base_formula, actual_formula, threshold_area=5, threshold_percentage=1
):
    """
    function that checks if 2 formulas are equal (True,False) and adds an Evaluation
    """
    if base_formula is None or actual_formula is None:
        return False, Evaluation.TO_CHECK_NO_PREDICTION_5
    od_alike = False
    if base_formula["reference_od"] is None and actual_formula["reference_od"] is None:
        od_alike = True
    elif base_formula["reference_od"] is None or actual_formula["reference_od"] is None:
        od_alike = False
    elif (
        abs(
            base_formula["reference_od"]["area"]
            - actual_formula["reference_od"]["area"]
        )
        * 100
        / base_formula["reference_od"]["area"]
    ) < threshold_percentage:
        od_alike = True

    if (
        base_formula["reference_features"].keys()
        == actual_formula["reference_features"].keys()
        and od_alike
    ):
        if base_formula["full"] and actual_formula["full"]:
            return True, Evaluation.EQUALITY_EQUAL_FORMULA_FULL_1

        equal_reference_features = True
        for key in base_formula["reference_features"].keys():
            if (
                (
                    base_formula["reference_features"][key]["full"]
                    == actual_formula["reference_features"][key]["full"]
                )
                or (
                    abs(
                        base_formula["reference_features"][key]["area"]
                        - actual_formula["reference_features"][key]["area"]
                    )
                    > threshold_area
                )
                or (
                    (
                        abs(
                            base_formula["reference_features"][key]["area"]
                            - actual_formula["reference_features"][key]["area"]
                        )
                        * 100
                        / base_formula["reference_features"][key]["area"]
                    )
                    > threshold_percentage
                )
            ):
                equal_reference_features = False
        if equal_reference_features:
            return True, Evaluation.EQUALITY_EQUAL_FORMULA_2
    if base_formula["full"] and actual_formula["full"] and od_alike:
        return True, Evaluation.EQUALITY_FULL_3
    return False, Evaluation.TO_CHECK_NO_PREDICTION_5



import json
import unittest

import numpy as np
from shapely import to_geojson
from shapely.geometry import shape

from brdr.aligner import Aligner
from brdr.enums import DiffMetric
from brdr.enums import GRBType
from brdr.enums import OpenbaarDomeinStrategy
from brdr.grb import GRBActualLoader
from brdr.loader import DictLoader
from brdr.utils import diffs_from_dict_series


class TestExamples(unittest.TestCase):

    def test_webservice_brdr(self):
        """
        Code used in brdr-webservice
        """
        contour = {
            "type": "MultiPolygon",
            "coordinates": [
                [
                    [
                        [174647.924166895216331, 170964.980246312363306],
                        [174693.204879119351972, 170943.531487890402786],
                        [174696.382472959638108, 170930.82111252922914],
                        [174678.905706838035258, 170901.428369506524177],
                        [174660.634542256389977, 170861.708446502889274],
                        [174641.56897921464406, 170820.399726579111302],
                        [174593.905071610264713, 170839.46528962085722],
                        [174614.559431572153699, 170881.568408004706725],
                        [174628.064205393398879, 170926.054721768799936],
                        [174647.924166895216331, 170964.980246312363306],
                    ]
                ]
            ],
        }
        referentielaag_type = GRBType.ADP
        openbaardomein_strategy = OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE

        aligner = Aligner(area_limit=100000)

        geometry = shape(contour)

        aligner.load_thematic_data(DictLoader({"input_id": geometry}))
        aligner.load_reference_data(
            GRBActualLoader(
                grb_type=referentielaag_type, partition=1000, aligner=aligner
            )
        )

        series = np.arange(0, 61, 1, dtype=float) / 10

        dict_series = aligner.process(
            relevant_distances=series,
            od_strategy=openbaardomein_strategy,
            threshold_overlap_percentage=50,
        )
        dict_diffs = diffs_from_dict_series(
            dict_series, aligner.dict_thematic, DiffMetric.CHANGES_AREA
        )

        dict_diffs = dict_diffs["input_id"]
        serial_dict = {}
        dict_results = dict_series["input_id"]
        for rel_dist, process_results in dict_results.items():
            serial_dict[rel_dist] = {
                "result": json.loads(to_geojson(dict_results[rel_dist]["result"])),
                "result_diff_min": json.loads(
                    to_geojson(dict_results[rel_dist]["result_diff_min"])
                ),
                "result_diff_plus": json.loads(
                    to_geojson(dict_results[rel_dist]["result_diff_plus"])
                ),
            }

        return {
            "series": serial_dict,
            "diffs": dict_diffs,
        }



import json
from datetime import date

import numpy as np

from brdr.aligner import Aligner
from brdr.constants import EVALUATION_FIELD_NAME
from brdr.enums import GRBType, AlignerResultType
from brdr.grb import GRBActualLoader
from brdr.grb import GRBFiscalParcelLoader
from brdr.grb import get_affected_by_grb_change
from brdr.loader import DictLoader, GeoJsonFileLoader

# Press the green button in the gutter to run the script.
if __name__ == "__main__":
    """
    EXAMPLE of the 'evaluate()-function of 'brdr': This function evaluates thematic objects with a former brdr_formula and compares them with an actual formula; and adds evaluation-properties to the result
    """
    # initiate a base Aligner, to align thematic objects on an older version of the parcels (year 2022)
    base_aligner = Aligner()
    # Load thematic data
    loader = GeoJsonFileLoader("themelayer.geojson", "theme_identifier")
    base_aligner.load_thematic_data(loader)
    base_year = "2022"
    name_formula = "base_formula"
    # Load reference data
    base_aligner.load_reference_data(
        GRBFiscalParcelLoader(year=base_year, aligner=base_aligner)
    )
    relevant_distance = 2
    # Align the thematic object on the parcelborders of 2022, to simulate a base-situation
    base_process_result = base_aligner.process(relevant_distance=2)

    # Collect the base-situation (base-geometries and the brdr_formula from that moment
    thematic_dict_formula = {}
    thematic_dict_result = {}
    for key in base_process_result:
        thematic_dict_result[key] = base_process_result[key][relevant_distance][
            "result"
        ]
        thematic_dict_formula[key] = {
            name_formula: json.dumps(
                base_aligner.get_brdr_formula(thematic_dict_result[key])
            )
        }
        print(key + ": " + thematic_dict_result[key].wkt)
        print(key + ": " + str(thematic_dict_formula[key]))

    # (OPTIONAL) Check for changes in the period 2022-now of the reference-parcels (GRB/Flanders-specific function)
    affected, unaffected = get_affected_by_grb_change(
        dict_thematic=thematic_dict_result,
        grb_type=GRBType.ADP,
        date_start=date(2022, 1, 1),
        date_end=date.today(),
        one_by_one=False,
    )
    if len(affected) == 0:
        print("No affected dicts")
        exit()
    print("Affected_IDs: " + str(affected))

    # Start an aligner to align thematic objects on the actual parcels
    actual_aligner = Aligner(relevant_distances=np.arange(0, 210, 10, dtype=int) / 100)
    # Load the thematic objects (aligned on 2022) and also give the brdr_formula from 2022 as property
    actual_aligner.load_thematic_data(
        DictLoader(
            data_dict=thematic_dict_result, data_dict_properties=thematic_dict_formula
        )
    )
    # Load reference data; the actual parcels
    actual_aligner.load_reference_data(
        GRBActualLoader(grb_type=GRBType.ADP, partition=1000, aligner=actual_aligner)
    )
    # Use the EVALUATE-function
    dict_evaluated, prop_dictionary = actual_aligner.evaluate(
        ids_to_evaluate=affected, base_formula_field=name_formula
    )

    # SHOW the EVALUATED results
    fc = actual_aligner.get_results_as_geojson(
        resulttype=AlignerResultType.EVALUATED_PREDICTIONS,
        formula=True,
        attributes=True,
    )
    print(fc["result"])

    for feature in fc["result"]["features"]:
        print(
            feature["properties"][actual_aligner.name_thematic_id]
            + ": "
            + feature["properties"][EVALUATION_FIELD_NAME]
        )



from brdr.aligner import Aligner
from brdr.enums import OpenbaarDomeinStrategy, AlignerResultType
from brdr.geometry_utils import geom_from_wkt
from brdr.loader import DictLoader

if __name__ == "__main__":
    """
    Example to load dat with a DictLoader, and also adding the properties to the result
    """
    # CREATE AN ALIGNER
    aligner = Aligner(crs="EPSG:31370", multi_as_single_modus=True)
    # ADD A THEMATIC POLYGON TO THEMATIC DICTIONARY and LOAD into Aligner
    id = 1
    thematic_dict = {id: geom_from_wkt("POLYGON ((0 0, 0 9, 5 10, 10 0, 0 0))")}
    # Add properties
    thematic_dict_properties = {
        id: {"propA": 1, "propB": 1.1, "propC": "dit is tekst", "propD": None}
    }
    loader = DictLoader(
        data_dict=thematic_dict, data_dict_properties=thematic_dict_properties
    )
    aligner.load_thematic_data(loader)
    # ADD A REFERENCE POLYGON TO REFERENCE DICTIONARY and LOAD into Aligner
    reference_dict = {100: geom_from_wkt("POLYGON ((0 1, 0 10,8 10,10 1,0 1))")}
    loader = DictLoader(reference_dict)
    aligner.load_reference_data(loader)
    # EXECUTE THE ALIGNMENT
    relevant_distance = 1
    process_result = aligner.process(
        relevant_distance=relevant_distance,
        od_strategy=OpenbaarDomeinStrategy.SNAP_SINGLE_SIDE,
        threshold_overlap_percentage=50,
    )
    # PRINT RESULTS IN WKT
    print("result: " + process_result[id][relevant_distance]["result"].wkt)
    print(
        "added area: " + process_result[id][relevant_distance]["result_diff_plus"].wkt
    )
    print(
        "removed area: " + process_result[id][relevant_distance]["result_diff_min"].wkt
    )
    fcs = aligner.get_results_as_geojson(
        resulttype=AlignerResultType.PROCESSRESULTS, formula=True, attributes=True
    )
    print(fcs["result"])



from brdr.aligner import Aligner
from brdr.enums import GRBType
from brdr.grb import GRBActualLoader
from brdr.loader import DictLoader, GeoJsonUrlLoader
from brdr.oe import get_oe_dict_by_ids


class TestExamples:

    def test_load_data(self):
        # EXAMPLE
        aligner = Aligner()

        dict_theme = get_oe_dict_by_ids([131635])
        thematic_loader = DictLoader(data_dict=dict_theme)
        reference_loader = GRBActualLoader(
            grb_type=GRBType.ADP, aligner=aligner, partition=1000
        )

        aligner.dict_thematic, props_thematic, thematic_source = (
            thematic_loader.load_data()
        )
        aligner.dict_reference, props_reference, reference_source = (
            reference_loader.load_data()
        )
        assert aligner.dict_reference is not None


def test_load_thematic_data_by_url(requests_mock, haspengouw_geojson):
    requests_mock.add(
        requests_mock.GET,
        "https://mock.com/haspengouw.geojson",
        json=haspengouw_geojson,
        status=200,
    )
    aligner = Aligner()
    aligner.load_thematic_data(
        GeoJsonUrlLoader("https://mock.com/haspengouw.geojson", "Id")
    )
    assert aligner.dict_thematic is not None


def test_load_reference_data_url(requests_mock, haspengouw_geojson):
    requests_mock.add(
        requests_mock.GET,
        "https://mock.com/haspengouw.geojson",
        json=haspengouw_geojson,
        status=200,
    )
    aligner = Aligner()

    aligner.load_reference_data(
        GeoJsonUrlLoader("https://mock.com/haspengouw.geojson", "Id")
    )

    assert aligner.dict_reference is not None



from enum import Enum
from enum import IntEnum


class OpenbaarDomeinStrategy(IntEnum):
    """
    Determines how the algorithm deals with parts of the geometry that are not on the
    reference layer. (=public domain in the case of plots as a reference layer).
    Different strategies have been developed:

    *   -1 (EXCLUDE): Completely exclude everything that is not on the reference layer
    *   0 (AS_IS): All parts that are not covered by the reference layer are added to
        the resulting geometry AS IS
    *   1 (SNAP_SINGLE_SIDE): Everything that falls within the relevant distance over
        the plot boundary is snapped to the plot. The outer boundary is not used.
    *   2 (SNAP_ALL_SIDE): Everything that falls within the relevant distance over the
        plot boundary is snapped to the plot. The outer boundary is used.
    *   3 (SNAP_FULL_AREA_SINGLE_SIDE): integrates entire Openbaar Domein within
        buffered original geometry
    *   4 (SNAP_FULL_AREA_ALL_SIDE): integrates all OD within inside-buffered
        original geometry. The first part is a copy of SNAP_ALL_SIDE
    *   5 (SNAP_SINGLE_SIDE_VARIANT_1): implementation variant 1 of strategy
        SNAP_SINGLE_SIDE
    *   6 (SNAP_SINGLE_SIDE_VARIANT_2): implementation variant 2 of strategy
        SNAP_SINGLE_SIDE

    """

    EXCLUDE = -1
    AS_IS = 0
    SNAP_SINGLE_SIDE = 1
    SNAP_ALL_SIDE = 2
    SNAP_FULL_AREA_SINGLE_SIDE = 3
    SNAP_FULL_AREA_ALL_SIDE = 4
    SNAP_SINGLE_SIDE_VARIANT_1 = 5
    SNAP_SINGLE_SIDE_VARIANT_2 = 6


class AlignerResultType(str, Enum):
    """
    2 Types of resulting dictionary are available in Aligner

    * PREDICTIONS = "predictions" (only the predicted versions for specific relevenat distances are returned)
    * PROCESSRESULTS = "processresults" (All versions of resulting geometries for all relevant distances are returned)
    """

    PREDICTIONS = "predictions"
    EVALUATED_PREDICTIONS = "evaluated_predictions"
    PROCESSRESULTS = "processresults"


class AlignerInputType(str, Enum):
    """
    2 Types of input dictionary are available in Aligner

    * THEMATIC = "thematic"
    * REFERENCE = "reference"
    """

    THEMATIC = "thematic"
    REFERENCE = "reference"


class GRBType(str, Enum):
    """
    Determines which GRB feature collection is used. Different types are available:

    * ADP: administrative plots
    * GBG: buildings on the ground
    * KNW: artworks
    """

    ADP = "adp"
    GBG = "gbg"
    KNW = "knw"


class DiffMetric(str, Enum):
    """
    Determines which metric is used to determine the difference between the thematic and
    reference data. Different metrics are available:

    * TOTAL_AREA: the total difference area
    * TOTAL_PERCENTAGE: the percentage of the total difference area
    * CHANGES_AREA: the sum of the negative and positive difference areas
    * CHANGES_PERCENTAGE: the percentage of the changed area
    """

    TOTAL_AREA = "total_area"
    TOTAL_PERCENTAGE = "total_percentage"
    CHANGES_AREA = "changes_area"
    CHANGES_PERCENTAGE = "changes_percentage"


class Evaluation(str, Enum):
    """
    Enum to evaluate an automatically updated geometry:

    EQUALITY_EQUAL_FORMULA_FULL_1 = "equality_equal_formula_full_1"
    EQUALITY_EQUAL_FORMULA_2 = "equality_equal_formula_2"
    EQUALITY_FULL_3 = "equality_full_3"
    TO_CHECK_PREDICTION_4 = "to_check_prediction_4"
    TO_CHECK_NO_PREDICTION_5 = "to_check_no_prediction_5"
    NO_CHANGE_6 = "no_change_6"
    """

    EQUALITY_EQUAL_FORMULA_FULL_1 = "equality_equal_formula_full_1"
    EQUALITY_EQUAL_FORMULA_2 = "equality_equal_formula_2"
    EQUALITY_FULL_3 = "equality_full_3"
    TO_CHECK_PREDICTION_4 = "to_check_prediction_4"
    TO_CHECK_NO_PREDICTION_5 = "to_check_no_prediction_5"
    NO_CHANGE_6 = "no_change_6"
